theory Matrix imports Util AbstractRewriting
begin


section {* matrices *}

subsection {* types and well-formedness of vectors / matrices *}

types 'a vec = "'a list"
types 'a mat = "'a vec list" (* list of column-vectors *)


(* well formed vector *)
definition wf_vec :: "nat \<Rightarrow> 'x vec \<Rightarrow> bool"
 where "wf_vec n x = (length x = n)"

(* well formed matrix w.r.t. number of rows and columns *)
definition wf_mat :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" where
 "wf_mat nr nc m = (length m = nc \<and> forall (wf_vec nr) m)"

subsection {* definitions / algorithms *}

text {* note that these algorithms are generic in all basic definitions / operations 
like 0 (ze) 1 (on) addition (pl) multiplication (ti) and in the dimension(s) of the matrix/vector.
Hence, many of these algorithms require these definitions/operations/sizes as arguments.
All indices start from 0.
*}

(* the 0 vector *)
definition vec0I :: "'a \<Rightarrow> nat \<Rightarrow> 'a vec" where 
 "vec0I ze n = replicate n ze"

(* the 0 matrix *)
definition mat0I :: "'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a mat" where
  "mat0I ze nr nc = replicate nc (vec0I ze nr)"

(* the i-th unit vector of size n *) 
definition vec1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a vec"
  where "vec1I ze on n i \<equiv> replicate i ze @ on # replicate (n - 1 - i) ze"

(* the 1 matrix *)
definition mat1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a mat"
  where "mat1I ze on n \<equiv> map (vec1I ze on n) [0 ..< n]"


(* vector addition *)
definition vec_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec" where 
 "vec_plusI pl v w = map (\<lambda> xy. pl (fst xy) (snd xy)) (zip v w)"

(* matrix addition *)
definition mat_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
 where "mat_plusI pl m1 m2 = map (\<lambda> uv. vec_plusI pl (fst uv) (snd uv)) (zip m1 m2)"

(* scalar product *)
definition scalar_prodI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a" where
 "scalar_prodI ze pl ti v w = foldr (\<lambda> (x,y) s. pl (ti x y) s) (zip v w) ze"

(* the m-th row of a matrix *)
definition row :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a vec"
where "row m i \<equiv> map (\<lambda> w. w ! i) m"

(* the m-th column of a matrix *)
definition col :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a vec"
where "col m i \<equiv> m ! i"

(* transposition of a matrix (number of rows of matrix has to be given since otherwise one 
   could not compute transpose [] which might be [] or [[]] or [[], []], or ...) *)
fun transpose :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
 where "transpose nr [] = replicate nr []"
     | "transpose nr (v # m) = map (\<lambda> (vi,mi). (vi # mi)) (zip v (transpose nr m))"

(* matrix-vector multiplication, assumes the transposed matrix is given *)
definition matT_vec_multI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
 where "matT_vec_multI ze pl ti m v = map (\<lambda> w. scalar_prodI ze pl ti w v) m"

(* matrix-matrix multiplication, number of rows of left matrix has to be given (as transpose is used) *)
definition mat_multI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat" 
where "mat_multI ze pl ti nr m1 m2 \<equiv> map (matT_vec_multI ze pl ti (transpose nr m1)) m2"

(* taking an only the n first elements of a vector *)
definition sub_vec :: "nat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
where "sub_vec = take"

(* taking only the upper left sub matrix *)
definition sub_mat :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "sub_mat nr nc m = map (sub_vec nr) (take nc m)"

(* comparison of vectors using >= componentwise *)
definition vec_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_geI ge v w \<equiv> forall (\<lambda> (a,b). ge a b) (zip v w)"

(* comparison of matrices using >= componentwise *)
definition mat_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_geI ge m1 m2 \<equiv> forall (\<lambda> (v,w). vec_geI ge v w) (zip m1 m2)"

(* demanding at least one strict decrease between two vectors *)
definition vec_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_pre_gtI gt v w \<equiv> any (\<lambda> (a,b). gt a b) (zip v w)"

(* demanding at least one strict decrease between two matrices *)
definition mat_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_pre_gtI gt m1 m2 \<equiv> any (\<lambda> (v,w). vec_pre_gtI gt v w) (zip m1 m2)"

(* strict comparison of matrices is done by demanding that all entries are weakly
   decreasing and that there is at least one entry in the upper left sub-matrices
   which strictly decreases *)      
definition mat_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_gtI ge gt sd m1 m2 \<equiv> mat_geI ge m1 m2 \<and> mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2)"

(* checking whether a matrix is monotone w.r.t. >. To this end, 
   it is ensured that all columns in the upper left sub-matrix have an entry 
   of at least 1 *)
definition mat_monoI :: "('a \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_monoI geq1 sd m = forall (any geq1) (sub_mat sd sd m)"


(* applying max0 compenentwise to vectors *)
definition vec_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
  where "vec_max0I = map"

(* applying max0 compenentwise to matrices *)
definition mat_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
  where "mat_max0I mz = map (vec_max0I mz)"

(* checking whether a matrix is arctic positive (first entry is arctic positive) *)
definition mat_arc_posI :: "('a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_arc_posI ap m \<equiv> ap (m ! 0 ! 0)"


subsection {* algorithms preserve well-formedness *}


lemma wf_vec0[simp] : "wf_vec nr (vec0I ze nr)"
  by (simp add: wf_vec_def vec0I_def)

lemma wf_mat0[simp] : "wf_mat nr nc (mat0I ze nr nc)"
  by (simp add: wf_mat_def mat0I_def, rule replicate_prop, simp)


lemma wf_vec1: assumes "i < nr" shows "wf_vec nr (vec1I ze on nr i)"
unfolding wf_vec_def vec1I_def using assms by auto

lemma wf_mat1: "wf_mat nr nr (mat1I ze on nr)"
unfolding wf_mat_def mat1I_def using wf_vec1 by auto

lemma wf_vec_plus : "\<lbrakk>wf_vec nr u; wf_vec nr v\<rbrakk> \<Longrightarrow> wf_vec nr (vec_plusI pl u v)"
using assms 
unfolding vec_plusI_def wf_vec_def
by auto

lemma wf_mat_plus : assumes "wf_mat nr nc m1" and "wf_mat nr nc m2" shows "wf_mat nr nc (mat_plusI pl m1 m2)"
using assms
unfolding wf_mat_def mat_plusI_def
proof (simp, induct nc arbitrary: m1 m2, simp)
  case (Suc nn)
  show ?case 
  proof (cases m1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons v1 mm1) note oCons = this
    with Suc have l1: "length mm1 = nn" by auto
    show ?thesis
    proof (cases m2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons v2 mm2)
      with Suc have l2: "length mm2 = nn" by auto
      show ?thesis by  (simp add: Cons oCons, rule conjI, rule wf_vec_plus, (simp add: Cons oCons Suc)+, rule Suc, auto simp: Cons oCons Suc l1 l2)
    qed
  qed
qed

lemma wf_row : assumes "wf_mat nr nc m"
  and "i < nr"
  shows "wf_vec nc (row m i)"
  using assms
  unfolding wf_vec_def row_def wf_mat_def
  by (auto simp: wf_vec_def) 

lemma wf_col : assumes "wf_mat nr nc m"
  and "i < nc"
  shows "wf_vec nr (col m i)"
  using assms
  unfolding wf_vec_def col_def wf_mat_def
  by (auto simp: wf_vec_def) 

lemma wf_transpose : assumes "wf_mat nr nc m"
  shows "wf_mat nc nr (transpose nr m)"
using assms 
proof (induct m arbitrary: nc, simp add: wf_mat_def wf_vec_def set_replicate_conv_if)
  case (Cons v m)
  from `wf_mat nr nc (v # m)` obtain ncc where nc: "nc = Suc ncc" by (cases nc, auto simp: wf_mat_def) 
  with Cons have wfRec: "wf_mat ncc nr (transpose nr m)" unfolding wf_mat_def by auto
  from `wf_mat nr nc (v # m)` have wfV: "length v = nr" unfolding wf_mat_def by (simp add:  wf_vec_def)
  show ?case unfolding wf_mat_def
  proof (rule conjI, simp add: wfV wfRec wf_mat_def wf_vec_def)
    show "min nr (length (transpose nr m)) = nr" using wfRec unfolding wf_mat_def by auto
  next
    show "forall (wf_vec nc) (transpose nr (v # m))"
    proof (simp add: wf_vec_def nc, rule ballI, simp only: split_paired_all)
      fix a b
      assume mem: "(a,b) \<in> set (zip v (transpose nr m))"
      from mem have "b \<in> set (transpose nr m)" by (rule set_zip_rightD)
      with wfRec have "length b = ncc" unfolding wf_mat_def using wf_vec_def[of ncc] by auto
      thus "length (split op # (a,b)) = Suc ncc" by auto
    qed
  qed
qed


lemma wf_matT_vec_multI : assumes "wf_mat nr nc m"
  shows "wf_vec nc (matT_vec_multI ze pl ti m v)"
  unfolding matT_vec_multI_def
  using assms
  unfolding wf_mat_def
  by (simp add: wf_vec_def)

lemma wf_mat_mult : assumes wf1: "wf_mat nr n m1"
  and wf2: "wf_mat n nc m2"
  shows "wf_mat nr nc (mat_multI ze pl ti nr m1 m2)"
using assms
unfolding wf_mat_def mat_multI_def by (auto simp: wf_matT_vec_multI[OF wf_transpose[OF wf1]])

lemma wf_vec_max0: assumes "wf_vec nr v" shows "wf_vec nr (vec_max0I mz v)"
using assms unfolding wf_vec_def vec_max0I_def by auto

lemma wf_mat_max0: assumes "wf_mat nr nc m" shows "wf_mat nr nc (mat_max0I mz m)"
using assms unfolding wf_mat_def mat_max0I_def
  by (auto simp: wf_vec_max0)

lemma wf_sub_vec: assumes "wf_vec nr v" and "sd \<le> nr" 
  shows "wf_vec sd (sub_vec sd v)"
using assms unfolding wf_vec_def sub_vec_def by auto

lemma wf_sub_mat: assumes wf: "wf_mat nr nc m" and sr: "sr \<le> nr" and sc: "sc \<le> nc"
  shows "wf_mat sr sc (sub_mat sr sc m)"
using assms in_set_takeD[of _ sc m] wf_sub_vec[OF _ sr] unfolding wf_mat_def sub_mat_def by auto


subsection {* properties of algorithms which do not depend on properties of type of matrix *}

lemma mat0_index: assumes "i < nc" and "j < nr"
  shows "mat0I ze nr nc ! i ! j = ze"
unfolding mat0I_def vec0I_def using assms by auto

lemma mat0_row: assumes "i < nr"
  shows "row (mat0I ze nr nc) i = vec0I ze nc"
unfolding row_def mat0I_def vec0I_def
using assms by auto

lemma mat0_col: assumes "i < nc"
  shows "col (mat0I ze nr nc) i = vec0I ze nr"
unfolding mat0I_def col_def
using assms by auto

lemma vec1_index : assumes i: "i < n"
  and j: "j < n"
  shows "vec1I ze on n i ! j = (if i = j then on else ze)" (is "_ = ?r")
unfolding vec1I_def
proof -
  let ?l = "replicate i ze @ on # replicate (n - 1 - i) ze"
  have len: "length ?l > i" by auto
  have len2: "length (replicate i ze @ on # []) > i" by auto
  show "?l ! j = ?r"
  proof (cases "j = i")
    case True
    thus ?thesis by (simp add: nth_append)
  next
    case False
    show ?thesis 
    proof (cases "j < i")
      case True
      thus ?thesis by (simp add: nth_append)
    next
      case False
      with `j \<noteq> i` have gt: "j > i" by auto
      from this have "\<exists> k. j = i + Suc k" by arith
      from this obtain k where k: "j = i + Suc k" by auto
      with j i show ?thesis by (simp add: nth_append)
    qed
  qed
qed


lemma col_transpose_is_row: 
  assumes wf: "wf_mat nr nc m"
  and i: "i < nr"
  shows "col (transpose nr m) i = row m i"
using wf 
proof (induct m arbitrary: nc , simp add: col_def row_def wf_mat_def i)
  case (Cons v m)
  from `wf_mat nr nc (v # m)` obtain ncc where nc: "nc = Suc ncc" and wf: "wf_mat nr ncc m"  by (cases nc, auto simp: wf_mat_def)
  from `wf_mat nr nc (v # m)` nc have lengths: "(\<forall> w \<in> set m. length w = nr) \<and> length v = nr \<and> length m = ncc" unfolding wf_mat_def by (auto simp: wf_vec_def)
  from wf Cons have colRec: "col (transpose nr m) i = row m i" by auto
  from wf have wf_trans: "wf_mat ncc nr (transpose nr m)" by (rule wf_transpose)
  hence lengths2: "(\<forall> w \<in> set (transpose nr m). length w = ncc) \<and> length (transpose nr m) = nr" unfolding wf_mat_def by (auto simp: wf_vec_def)
  show ?case proof (rule nth_equalityI, simp add: col_def row_def lengths lengths2 i, intro allI impI)
    fix j
    have simpme: "transpose nr m ! i = row m i" using colRec unfolding col_def by auto
    assume "j < length (col (transpose nr (v # m)) i)"
    hence "j < Suc ncc" by (simp add: col_def lengths2 lengths i) 
    thus "col (transpose nr (v # m)) i ! j = row (v # m) i ! j"
      by (cases j, simp add: row_def col_def i lengths lengths2, simp add: row_def col_def i lengths lengths2 simpme)
  qed
qed

lemma mat_col_eq:
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  shows "(\<forall> i < nc. col m1 i = col m2 i) = (m1 = m2)" (is "?l = ?r")
proof
  assume ?r thus ?l by auto
next
  assume ?l show ?r 
  proof (rule nth_equalityI)
    show "length m1 = length m2" using wf1 wf2 unfolding wf_mat_def by auto
  next
    from `?l` show "\<forall> i < length m1. m1 ! i = m2 ! i" using wf1 unfolding col_def wf_mat_def by auto
  qed
qed

lemma mat_eq_index:
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  shows "(m1 = m2) = (\<forall> i < nc. \<forall> j < nr. m1 ! i ! j = m2 ! i ! j)" (is "?l = ?r")
proof 
  assume ?l thus ?r by auto
next
  assume ?r show ?l
  proof (simp only: mat_col_eq[OF wf1 wf2,symmetric], unfold col_def, intro allI impI)
    fix i
    assume i: "i < nc"
    show "m1 ! i = m2 ! i"
    proof (rule nth_equalityI)      
      show "length (m1 ! i)  = length (m2 ! i)" using wf1 wf2 i unfolding wf_mat_def by (auto simp: wf_vec_def)
    next
      from `?r` i show "\<forall> j < length (m1 ! i). m1 ! i ! j = m2 ! i ! j" using wf1 wf2 unfolding wf_mat_def by (auto simp: wf_vec_def)
    qed
  qed
qed

lemma vec_index_eq: 
  assumes wf1: "wf_vec n v1"
  and wf2: "wf_vec n v2"
  shows "(v1 = v2) = (\<forall> i < n. v1 ! i = v2 ! i)" (is "?l = ?r")
proof
  assume ?l thus ?r by auto
next
  assume ?r show ?l 
  proof (rule nth_equalityI)
    from wf1 wf2 show "length v1 = length v2" unfolding wf_vec_def by simp
  next
    from `?r` wf1 show "\<forall> i < length v1. v1 ! i = v2 ! i" unfolding wf_vec_def by simp
  qed
qed

lemma row_col : assumes "wf_mat nr nc m"  
  and "i < nr" and "j < nc"
  shows "row m i ! j = col m j ! i"
using assms unfolding wf_mat_def row_def col_def
  by auto

lemma mat_row_eq: 
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  shows "(m1 = m2) = (\<forall> i < nr. row m1 i = row m2 i)" (is "?l = ?r")
proof 
  assume ?l thus ?r by auto
next
  assume ?r show ?l
  proof (rule nth_equalityI)
    show "length m1 = length m2" using wf1 wf2 unfolding wf_mat_def by auto
  next
    show "\<forall> i < length m1. m1 ! i = m2 ! i"
    proof (intro allI impI)
      fix i
      assume i: "i < length m1"
      show "m1 ! i = m2 ! i"
      proof (rule nth_equalityI)
	show "length (m1 ! i) = length (m2 ! i)" using wf1 wf2 i unfolding wf_mat_def by (auto simp: wf_vec_def)
      next
	show "\<forall> j < length (m1 ! i). m1 ! i ! j = m2 ! i ! j" 
	proof (intro allI impI)
	  fix j
	  assume j: "j < length (m1 ! i)"
	  from i j wf1 have i1: "i < nc" and j1: "j < nr" unfolding wf_mat_def by (auto simp: wf_vec_def)
	  from `?r` j1 have "col m1 i ! j = col m2 i ! j"
	    by (simp add: row_col[OF wf1 j1 i1, symmetric] row_col[OF wf2 j1 i1, symmetric])
	  thus "m1 ! i ! j = m2 ! i ! j" unfolding col_def .
	qed
      qed
    qed
  qed
qed

lemma row_transpose_is_col:   assumes wf: "wf_mat nr nc m"
  and i: "i < nc"
  shows "row (transpose nr m) i = col m i"
proof -
  have len: "length (row (transpose nr m) i) = length (col m i)"
    using wf_transpose[OF wf]  wf i  unfolding row_def col_def wf_mat_def by (auto simp: wf_vec_def)
  show ?thesis 
  proof (rule nth_equalityI[OF len], intro allI impI)
    fix j
    assume "j < length (row (transpose nr m) i)"
    hence j: "j < nr" using wf_transpose[OF wf] wf i unfolding row_def col_def wf_mat_def by (auto simp: wf_vec_def)
    show "row (transpose nr m) i ! j = col m i ! j"
      by (simp only: row_col[OF wf_transpose[OF wf] i j],
	simp only: col_transpose_is_row[OF wf j],
	simp only: row_col[OF wf j i])
  qed
qed


lemma matT_vec_mult_to_scalar : 
  assumes "wf_mat nr nc m"
  and "wf_vec nr v"
  and "i < nc"
  shows "matT_vec_multI ze pl ti m v ! i = scalar_prodI ze pl ti (col m i) v"
unfolding matT_vec_multI_def using assms unfolding wf_mat_def col_def by (auto simp: wf_vec_def)

lemma mat_vec_mult_to_scalar : 
  assumes wf: "wf_mat nr nc m"
  and wfV: "wf_vec nc v"
  and i: "i < nr"
  shows "matT_vec_multI ze pl ti (transpose nr m) v ! i = scalar_prodI ze pl ti (row m i) v"
by (simp only:matT_vec_mult_to_scalar[OF wf_transpose[OF wf] wfV i],
  simp only: col_transpose_is_row[OF wf i])

lemma mat_mult_to_scalar :
  assumes wf1: "wf_mat nr n m1"
  and wf2: "wf_mat n nc m2"
  and i: "i < nr"
  and j: "j < nc"
  shows "mat_multI ze pl ti nr m1 m2 ! j ! i = scalar_prodI ze pl ti (row m1 i) (col m2 j)"
proof -
  have jlen: "j < length m2" using wf2 j unfolding wf_mat_def by auto
  have wfj: "wf_vec n (m2 ! j)" using jlen j wf2 unfolding wf_mat_def by auto
  show ?thesis 
    unfolding mat_multI_def
    by (simp add: jlen, simp only: mat_vec_mult_to_scalar[OF wf1 wfj i], unfold col_def, simp)
qed

lemma col_mat_mult_to_scalar :
  assumes wf1: "wf_mat nr n m1"
  and wf2: "wf_mat n nc m2"
  and j: "j < nc"
  shows "col (mat_multI ze pl ti nr m1 m2) j = map (\<lambda> i. scalar_prodI ze pl ti (row m1 i) (col m2 j)) [0 ..< nr]" (is "col ?l j = ?r")
proof - 
  have wf12: "wf_mat nr nc ?l" by (rule wf_mat_mult[OF wf1 wf2])
  have len: "length (col ?l j) = length ?r" and nr: "length (col ?l j) = nr" using wf1 wf2 wf12 j unfolding wf_mat_def col_def by (auto simp: wf_vec_def) 
  show ?thesis by (rule nth_equalityI[OF len], simp add: j nr, intro allI impI, unfold col_def, simp only:
    mat_mult_to_scalar[OF wf1 wf2 _ j], simp add: col_def)
qed

lemma row_mat_mult_to_scalar :
  assumes wf1: "wf_mat nr n m1"
  and wf2: "wf_mat n nc m2"
  and i: "i < nr"
  shows "row (mat_multI ze pl ti nr m1 m2) i = map (\<lambda> j. scalar_prodI ze pl ti (row m1 i) (col m2 j)) [0 ..< nc]" (is "row ?l i = ?r")
proof - 
  have wf12: "wf_mat nr nc ?l" by (rule wf_mat_mult[OF wf1 wf2])
  hence lenL: "length ?l = nc" unfolding wf_mat_def by simp
  have len: "length (row ?l i) = length ?r" and nc: "length (row ?l i) = nc" using wf1 wf2 wf12 i unfolding wf_mat_def row_def by (auto simp: wf_vec_def) 
  show ?thesis by (rule nth_equalityI[OF len], simp add: i nc, intro allI impI, unfold row_def, simp add: lenL, simp only: 
    mat_mult_to_scalar[OF wf1 wf2 i], simp add: row_def)
qed

lemma scalar_prod_cons : 
  "scalar_prodI ze pl ti (a # as) (b # bs) = pl (ti a b) (scalar_prodI ze pl ti as bs)"
unfolding scalar_prodI_def by auto


lemma vec_plus_index : 
  assumes wf1: "wf_vec nr v1"
  and wf2: "wf_vec nr v2"
  and i: "i < nr"
  shows "vec_plusI pl v1 v2 ! i = pl (v1 ! i)  (v2 ! i)"
using wf1 wf2 i
unfolding wf_vec_def vec_plusI_def
proof (induct v1 arbitrary: i v2 nr, simp)
  case (Cons a v11)
  from Cons obtain b v22 where v2: "v2 = b # v22" by (cases v2, auto)
  from v2 Cons obtain nrr where nr: "nr = Suc nrr" by (force)
  from Cons show ?case
    by (cases i, simp add: v2, auto simp: v2 nr)
qed

lemma mat_plus_index : 
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  and i: "i < nc"
  and j: "j < nr"
  shows "mat_plusI pl m1 m2 ! i ! j = pl (m1 ! i ! j) (m2 ! i ! j)"
using wf1 wf2 i
unfolding mat_plusI_def wf_mat_def 
proof (simp, induct m1 arbitrary: m2 i nc, simp)
  case (Cons v1 m11)
  from Cons obtain v2 m22 where m2: "m2 = v2 # m22" by (cases m2, auto)
  from m2 Cons obtain ncc where nc: "nc = Suc ncc" by force
  show ?case
  proof (cases i, simp add: m2, rule vec_plus_index[where nr = nr], (auto simp: Cons j m2)[3])
    case (Suc ii)
    with Cons show ?thesis using m2 nc by auto
  qed
qed

lemma col_mat_plus : assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  and i: "i < nc"
  shows "col (mat_plusI pl m1 m2) i = vec_plusI pl (col m1 i) (col m2 i)"
using assms
unfolding mat_plusI_def col_def wf_mat_def
proof (induct m1 arbitrary: m2 nc i, simp)
  case (Cons v m1)
  from Cons obtain v2 m22 where m2: "m2 = v2 # m22" by (cases m2, auto)
  from m2 Cons obtain ncc where nc: "nc = Suc ncc" by force
  show ?case
  proof (cases i, simp add: m2)
    case (Suc ii)
    with Cons show ?thesis using m2 nc by auto
  qed
qed

lemma transpose_index: assumes wf: "wf_mat nr nc m"
  and i: "i < nr"
  and j: "j < nc"
  shows "transpose nr m ! i ! j = m ! j ! i"
proof -
  have "transpose nr m ! i ! j = col (transpose nr m) i ! j" unfolding col_def by simp
  also have "\<dots> = row m i ! j" using col_transpose_is_row[OF wf i] by simp
  also have "\<dots> = m ! j ! i" unfolding row_def using wf j unfolding wf_mat_def by (auto simp: wf_vec_def)
  finally show ?thesis . 
qed

lemma transpose_mat_plus: assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  shows "transpose nr (mat_plusI pl m1 m2) = mat_plusI pl (transpose nr m1) (transpose nr m2)"
proof - 
  let ?m12 = "mat_plusI pl m1 m2"
  let ?t1 = "transpose nr m1"
  let ?t2 = "transpose nr m2"
  from wf_mat_plus[OF wf1 wf2] have wf12: "wf_mat nr nc ?m12" .
  from wf_transpose[OF wf1] have wft1: "wf_mat nc nr ?t1" .
  from wf_transpose[OF wf2] have wft2: "wf_mat nc nr ?t2" .
  show ?thesis 
  proof (simp only: mat_eq_index[OF wf_transpose[OF wf12] wf_mat_plus[OF wft1 wft2]], intro allI impI)
    fix i j
    assume i: "i < nr" and j: "j < nc"
    show "transpose nr ?m12 ! i ! j = mat_plusI pl ?t1 ?t2 ! i ! j"      
      by (simp only: transpose_index[OF wf12 i j],
	simp only: mat_plus_index[OF wft1 wft2 i j],
	simp only: mat_plus_index[OF wf1 wf2 j i],
	simp only: transpose_index[OF wf1 i j],
	simp only: transpose_index[OF wf2 i j])
  qed
qed
      

lemma row_mat_plus : assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  and i: "i < nr"
  shows "row (mat_plusI pl m1 m2) i = vec_plusI pl (row m1 i) (row m2 i)"
  by (
    simp only: col_transpose_is_row[OF wf_mat_plus[OF wf1 wf2] i, symmetric], 
    simp only: transpose_mat_plus[OF wf1 wf2],
    simp only: col_mat_plus[OF wf_transpose[OF wf1] wf_transpose[OF wf2] i],
    simp only: col_transpose_is_row[OF wf1 i],
    simp only: col_transpose_is_row[OF wf2 i])


lemma col_mat1: assumes "i < nr"
  shows "col (mat1I ze on nr) i = vec1I ze on nr i"
unfolding mat1I_def col_def using assms by auto


lemma mat1_index: assumes i: "i < n" and j: "j < n"
  shows "mat1I ze on n ! i ! j = (if i = j then on else ze)"
  by (simp add: col_mat1[OF i, simplified col_def] vec1_index[OF i j])

lemma transpose_mat1: "transpose nr (mat1I ze on nr) = (mat1I ze on nr)"
proof (simp only: mat_eq_index[OF wf_transpose[OF wf_mat1] wf_mat1], intro impI allI)
  fix i j
  assume i: "i < nr" and j: "j < nr"
  let ?I = "mat1I ze on nr"
  show "transpose nr ?I ! i ! j = ?I ! i ! j"
    by  (simp only: col_def[symmetric], 
      simp only: col_mat1[OF i],
      simp only: row_col[OF wf_transpose[OF wf_mat1] j i,symmetric],
      simp only: row_transpose_is_col[OF wf_mat1 j],
      simp only: col_mat1[OF j],
      simp only: vec1_index[OF i j],
      simp only: vec1_index[OF j i], simp)
qed

lemma row_mat1: assumes i: "i < nr"
  shows "row (mat1I ze on nr) i = vec1I ze on nr i"
by (simp only: col_transpose_is_row[OF wf_mat1 i, symmetric],
  simp only: transpose_mat1,
  simp only: col_mat1[OF i])

lemma vec_ge_index : assumes "wf_vec nr v1" 
  and "wf_vec nr v2"
  shows "vec_geI ge v1 v2 = (\<forall> i < nr. ge (v1 ! i) (v2 ! i))"
using assms
unfolding wf_vec_def vec_geI_def 
proof (simp, induct nr arbitrary: v1 v2, simp)
  case (Suc nrr)
  from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and lw1: "length w1 = nrr" by (cases v1, auto)
  from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and lw2: "length w2 = nrr" by (cases v2, auto)
  have rec: "(\<forall> a \<in> set (zip w1 w2). split ge a) = (\<forall> i < nrr. ge (w1 ! i) (w2 ! i))"
    by (rule Suc, auto simp: Suc lw1 lw2)
  show ?case (is "?l = ?r")
  proof (rule iffI)
    assume ?r
    thus ?l using Suc v1 v2 by auto
  next
    assume ?l
    show ?r
    proof (intro allI impI)
      fix i
      assume i: "i < Suc nrr"
      show "ge (v1 ! i) (v2 ! i)"
	using `?l` v1 v2 i
	by (cases i, auto simp: rec)
    qed
  qed
qed

lemma vec_pre_gt_index : assumes "wf_vec nr v1" 
  and "wf_vec nr v2"
  shows "vec_pre_gtI gt v1 v2 = (\<exists> i < nr. gt (v1 ! i) (v2 ! i))"
using assms[simplified wf_vec_def, symmetric]
by (simp only: Not_eq_iff[symmetric, of "vec_pre_gtI gt v1 v2"], unfold vec_pre_gtI_def, simp add: set_zip, auto) 
    
lemma mat_ge_index : assumes "wf_mat nr nc m1" 
  and "wf_mat nr nc m2"
  shows "mat_geI ge m1 m2 = (\<forall> i < nc. \<forall> j < nr. ge (m1 ! i ! j) (m2 ! i ! j))"
using assms
unfolding wf_mat_def mat_geI_def
proof (simp, induct nc arbitrary: m1 m2, simp)
  case (Suc ncc)
  from Suc obtain v1 mm1 where m1: "m1 = v1 # mm1" and lm1: "length mm1 = ncc \<and> (\<forall> a \<in> set mm1. wf_vec nr a)" by (cases m1, auto)
  from Suc obtain v2 mm2 where m2: "m2 = v2 # mm2" and lm2: "length mm2 = ncc \<and> (\<forall> a \<in> set mm2. wf_vec nr a)" by (cases m2, auto)
  from Suc m1 have wf1: "wf_vec nr v1" by simp
  from Suc m2 have wf2: "wf_vec nr v2" by simp
  have rec: "(\<forall> a \<in> set (zip mm1 mm2). split (vec_geI ge) a) = (\<forall> i < ncc. \<forall> j < nr. ge (mm1 ! i ! j) (mm2 ! i ! j))"
    by (rule Suc, auto simp: Suc lm1 lm2)
  show ?case (is "?l = ?r")
  proof (rule iffI)
    assume ?r
    thus ?l using m1 m2 lm1 lm2 rec vec_ge_index[OF wf1 wf2] by auto
  next
    assume ?l
    hence ge: "vec_geI ge v1 v2" and "\<forall> a \<in> set (zip mm1 mm2). split (vec_geI ge) a" using m1 m2 by auto
    with rec have ge2: " (\<forall>i<ncc. \<forall>j<nr. ge (mm1 ! i ! j) (mm2 ! i ! j))" by simp
    show ?r
    proof (rule allI, rule impI)
      fix i 
      assume i: "i < Suc ncc" 
      show "\<forall> j < nr. ge (m1 ! i ! j) (m2 ! i ! j)"
      proof (cases i, simp add: m1 m2, simp only: vec_ge_index[OF wf1 wf2, symmetric], rule ge)
	case (Suc ii)	
	with i have " ii < ncc" by simp
	with Suc 
	show ?thesis by (simp add: m1 m2, simp add: ge2)
      qed
    qed
  qed
qed

lemma mat_pre_gt_index : assumes "wf_mat nr nc m1"
  and "wf_mat nr nc m2"
  shows "mat_pre_gtI gt m1 m2 = (\<exists> i < nc. \<exists> j < nr. gt (m1 ! i ! j) (m2 ! i ! j))"
proof -
  from assms have l1: "nc = length m1" and l2: "length m2 = length m1" 
    and vl1: "\<forall> i < nc. wf_vec nr (m1 ! i)" and vl2: "\<forall> i < nc. wf_vec nr (m2 ! i)" unfolding wf_mat_def by auto
  show ?thesis
  proof (simp only: Not_eq_iff[symmetric, of "mat_pre_gtI gt m1 m2"], unfold mat_pre_gtI_def, simp, simp only: set_zip l2 min_max.inf_idem)
    have "(\<forall> x \<in> {(m1 ! i, m2 ! i) | i. i < length m1}. \<not> split (vec_pre_gtI gt) x) = (\<forall> i < nc. \<not> vec_pre_gtI gt (m1 ! i) (m2 ! i))" (is "?l = _")
      by (auto simp: l1)
    also have "\<dots> = (\<forall> i < nc. \<forall> j < nr. \<not> gt (m1 ! i ! j) (m2 ! i ! j))" (is "_ = ?r")
    proof -
      let ?l = "\<lambda> i. \<not> vec_pre_gtI gt (m1 ! i) (m2 ! i)"
      let ?r = "\<lambda> i. \<forall> j < nr. \<not> gt (m1 ! i ! j) (m2 ! i ! j)"
      have "\<forall> i < nc. ?l i = ?r i"
      proof (intro allI impI)
	fix i
	assume i: "i < nc"
	show "?l i = ?r i" using vec_pre_gt_index[OF mp[OF spec[OF vl1] i] mp[OF spec[OF vl2] i]] by auto
      qed	  
      thus ?thesis by blast
    qed
    finally show "?l = ?r" .
  qed
qed

lemma sub_mat_index:
  assumes wf: "wf_mat nr nc m"
  and sr: "sr \<le> nr"
  and sc: "sc \<le> nc"
  and j: "j < sr"
  and i: "i < sc"
  shows "sub_mat sr sc m ! i ! j = m ! i ! j"
proof -
  from assms have im: "i < length m" unfolding wf_mat_def by auto
  from assms have jm: "j < length (m ! i)" unfolding wf_mat_def by (auto simp: wf_vec_def)
  have "sub_mat sr sc m ! i ! j = map (take sr) (take sc m) ! i ! j"
    unfolding sub_mat_def sub_vec_def by auto
  also have "\<dots> = take sr (m ! i) ! j" using i im by auto
  also have "\<dots> = m ! i ! j" using j jm by auto
  finally show ?thesis .
qed

lemma mat_gt_index : assumes wf1: "wf_mat n n m1"
  and wf2: "wf_mat n n m2"
  and sd: "sd \<le> n"
  shows "mat_gtI ge gt sd m1 m2 = (mat_geI ge m1 m2 \<and> (\<exists> i < sd. \<exists> j < sd. gt (m1 ! i ! j) (m2 ! i ! j)))"
proof -
  have id: "mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2) = (\<exists> i < sd. \<exists> j < sd. gt (m1 ! i ! j) (m2 ! i ! j))" (is "?l = ?r")
  proof -
    have "?l = (\<exists> i < sd. \<exists> j < sd. gt (sub_mat sd sd m1 ! i ! j) (sub_mat sd sd m2 ! i ! j))"
      by (simp only: mat_pre_gt_index[OF wf_sub_mat[OF wf1 sd sd] wf_sub_mat[OF wf2 sd sd]])
    also have "\<dots> = ?r" by (simp only: Not_eq_iff[symmetric, of _ ?r], auto simp: sub_mat_index[OF wf1 sd sd] sub_mat_index[OF wf2 sd sd])
    finally show "?l = ?r" .
  qed
  thus ?thesis unfolding mat_gtI_def by auto
qed


lemma mat_mono_index : assumes wf: "wf_mat n n m"
  and sd: "sd \<le> n" 
  shows "mat_monoI geq1 sd m = (\<forall> i < sd. \<exists> j < sd. geq1 (m ! i ! j))"
proof -
  from wf sd have n: "n = length m" and sd: "sd \<le> length m"  and v: "\<forall> v \<in> set m. length v = n" unfolding wf_mat_def by (auto simp: wf_vec_def)
  have "set (take sd m) \<subseteq> set m" by (rule set_take_subset)
  with v have v: "\<forall> v \<in> set (take sd m). length v = n" by auto
  show ?thesis unfolding mat_monoI_def sub_mat_def sub_vec_def 
  proof (simp add: any_set_conv)
    show "(\<forall> x \<in> set (take sd m). \<exists> x \<in> set (take sd x). geq1 x) = (\<forall> i < sd. \<exists> j < sd. geq1 (m ! i ! j))"
      (is "?l = ?r")
    proof
      assume ?l
      show ?r
      proof (intro allI impI)
        fix i
        assume i: "i < sd" with sd have im: "i < length m" by auto
        have mi: "m ! i \<in> set (take sd m)" by (simp add: set_conv_nth, rule exI[of _ i],
          simp add: i im)        
        with `?l` have "\<exists> x \<in> set (take sd (m ! i)). geq1 x" by auto
        then obtain x where x: "x \<in> set (take sd (m ! i))" and geq: "geq1 x" by auto
        from v mi sd n have sdmi: "sd \<le> length (m ! i)" by auto
        from x[simplified set_conv_nth] obtain j where x: "x = take sd (m ! i) ! j" and 
          j: "j < length (take sd (m ! i))" by auto
        from j sdmi have j: "j < sd" by auto
        with x sdmi have x: "x = m ! i ! j" by auto
        from j geq x show "\<exists> j < sd. geq1 (m ! i ! j)" by auto
      qed
    next
      assume ?r
      show ?l
      proof (intro ballI)
        fix v
        assume vm: "v \<in> set (take sd m)"
        then obtain i where i: "i < sd" and vmi: "v = m ! i" by (simp add: set_conv_nth, auto)        
        with `?r` obtain j where j: "j < sd" and geq: "geq1 (m ! i ! j)" by auto
        from v vm sd n have sdv: "sd \<le> length v" by auto
        with j have j2: "j < length v" by auto
        have "v ! j \<in> set (take sd v)" by (simp add: set_conv_nth, rule exI[of _ j], simp add: j sdv j2)
        hence "m ! i ! j \<in> set (take sd v)" by (simp add: vmi)
        with geq show "\<exists> x \<in> set (take sd v). geq1 x" by auto
      qed
    qed
  qed
qed




subsection {* lemmas requiring properties of plus, times, ... *}

context plus
begin

abbreviation vec_plus :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
where "vec_plus \<equiv> vec_plusI plus"

abbreviation mat_plus :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "mat_plus \<equiv> mat_plusI plus"
end

context semigroup_add
begin
lemma vec_plus_assoc: assumes wf_u: "wf_vec nr u" and wf_v: "wf_vec nr v" and wf_w: "wf_vec nr w"
 shows "vec_plus u (vec_plus v w) = vec_plus (vec_plus u v) w" (is "?l = ?r")
proof -
  from wf_v wf_w have wf_vw: "wf_vec nr (vec_plus v w)" by (simp add: wf_vec_plus)
  from wf_u wf_v have wf_uv: "wf_vec nr (vec_plus u v)" by (simp add: wf_vec_plus)
  from assms have wf_l: "wf_vec nr ?l" by (simp add: wf_vec_plus)
  from assms have wf_r: "wf_vec nr ?r" by (simp add: wf_vec_plus)
  show ?thesis by (simp only: vec_index_eq[OF wf_l wf_r], intro allI impI,
    simp only: vec_plus_index[OF wf_u wf_vw],
    simp only: vec_plus_index[OF wf_v wf_w],
    simp only: vec_plus_index[OF wf_uv wf_w],
    simp only: vec_plus_index[OF wf_u wf_v],
    simp only: add_assoc)
qed

lemma mat_plus_assoc: assumes wf_1: "wf_mat nr nc m1" and wf_2: "wf_mat nr nc m2" and wf_3: "wf_mat nr nc m3"
  shows "mat_plus m1 (mat_plus m2 m3) = mat_plus (mat_plus m1 m2) m3" (is "?l = ?r")
proof -
  from wf_2 wf_3 have wf_23: "wf_mat nr nc (mat_plus m2 m3)" by (simp add: wf_mat_plus)
  from wf_1 wf_2 have wf_12: "wf_mat nr nc (mat_plus m1 m2)" by (simp add: wf_mat_plus)
  from assms have wf_l: "wf_mat nr nc ?l" by (simp add: wf_mat_plus)
  from assms have wf_r: "wf_mat nr nc ?r" by (simp add: wf_mat_plus)
  show ?thesis by (simp only: mat_eq_index[OF wf_l wf_r], intro allI impI,
    simp only: mat_plus_index[OF wf_1 wf_23],
    simp only: mat_plus_index[OF wf_2 wf_3],
    simp only: mat_plus_index[OF wf_12 wf_3],
    simp only: mat_plus_index[OF wf_1 wf_2],
    simp only: add_assoc)
qed
end

context ab_semigroup_add
begin
lemma vec_plus_comm : "vec_plus x y = vec_plus y x"
unfolding vec_plusI_def
proof (induct x arbitrary: y, simp)
  case (Cons a x)
  thus ?case 
    by (cases y, simp, simp add: add_commute) 
qed


lemma mat_plus_comm : "mat_plus m1 m2 = mat_plus m2 m1"
unfolding mat_plusI_def
proof (induct m1 arbitrary: m2, simp)
  case (Cons v m1) note oCons = this
  thus ?case
  proof (cases m2, simp)
    case (Cons w m2a)
    hence "mat_plus (v # m1) m2 = vec_plus v w # mat_plus m1 m2a" by (auto simp: mat_plusI_def)
    also have "\<dots> = vec_plus w v # mat_plus m1 m2a" using vec_plus_comm by auto
    finally show ?thesis using Cons oCons by (auto simp: mat_plusI_def)
  qed
qed
end

context zero
begin
abbreviation vec0 :: "nat \<Rightarrow> 'a vec"
where "vec0 \<equiv> vec0I zero"

abbreviation mat0 :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat"
where "mat0 \<equiv> mat0I zero"
end

context monoid_add
begin
lemma vec0_plus[simp]: assumes "wf_vec nr u" shows "vec_plus (vec0 nr) u = u"
using assms
unfolding wf_vec_def vec_plusI_def vec0I_def
proof (induct nr arbitrary: u, simp)
 case (Suc nn) thus ?case by (cases u, auto)
qed

lemma plus_vec0[simp]: assumes "wf_vec nr u" shows "vec_plus u (vec0 nr) = u"
using assms
unfolding wf_vec_def vec_plusI_def vec0I_def
proof (induct nr arbitrary: u, simp)
 case (Suc nn) thus ?case by (cases u, auto)
qed

lemma plus_mat0[simp]: assumes "wf_mat nr nc m" shows "mat_plus m (mat0 nr nc) = m"
using assms 
unfolding wf_mat_def 
proof (induct nc arbitrary: m, simp add: mat_plusI_def mat0I_def)
  case (Suc nn) 
  thus ?case proof (cases m, simp)
    case (Cons v mm)
    with Suc have wf: "wf_vec nr v" by auto
    from Cons Suc have "mat_plus m (mat0 nr (Suc nn)) = vec_plus v (vec0 nr) # mat_plus mm (mat0 nr nn)" by (auto simp: mat_plusI_def mat0I_def)
    also have "\<dots> = vec_plus v (vec0 nr) # mm" using Suc Cons by auto
    also have "\<dots> = v # mm" by (simp only: plus_vec0 wf)
    finally show ?thesis using Cons by auto
  qed
qed
end

context semiring_0
begin
abbreviation scalar_prod :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a"
where "scalar_prod \<equiv> scalar_prodI zero plus times"

abbreviation mat_mult :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "mat_mult \<equiv> mat_multI zero plus times"


lemma scalar_prod_last: assumes "length v1 = length v2" 
  shows "scalar_prod (v1 @ [x1]) (v2 @ [x2]) = x1 * x2 + scalar_prod v1 v2"
using assms 
proof (induct v1 arbitrary: v2, simp add: scalar_prodI_def)
  case (Cons y1 w1)
  from Cons(2) obtain y2 w2 where v2: "v2 = Cons y2 w2" and len: "length w1 = length w2" by (cases v2, auto)
  from Cons(1)[OF len] have rec: "scalar_prod (w1 @ [x1]) (w2 @ [x2]) = x1 * x2 + scalar_prod w1 w2" .
  have "scalar_prod ((y1 # w1) @ [x1]) (v2 @ [x2]) = 
    (y1 * y2 + x1 * x2) + scalar_prod w1 w2" by (simp add: scalar_prod_cons v2 rec add_assoc)
  also have "\<dots> = (x1 * x2 + y1 * y2) + scalar_prod w1 w2" using add_commute[of "x1 * x2"] by simp
  also have "\<dots> = x1 * x2 + (scalar_prod (y1 # w1) v2)" by (simp add: add_assoc scalar_prod_cons v2)
  finally show ?case .
qed

lemma scalar_product_assoc: 
  assumes wfm: "wf_mat nr nc m"
  and wfr: "wf_vec nr r"
  and wfc: "wf_vec nc c"
  shows "scalar_prod (map (\<lambda>k. scalar_prod r (col m k)) [0..<nc]) c = scalar_prod r (map (\<lambda>k. scalar_prod (row m k) c) [0..<nr])"
using wfm wfc
unfolding col_def 
proof (induct m arbitrary: nc c)
  case Nil
  hence "nc = 0" unfolding wf_mat_def by (auto)
  with Nil show ?case 
  proof (simp add: scalar_prodI_def row_def)
    from wfr have nr: "nr = length r" unfolding wf_vec_def by auto
    let ?term = "\<lambda> r :: 'a vec. zip r (map (\<lambda> k. zero) [0..<length r])"
    let ?fun = "\<lambda> (x,y). plus (times x y)"
    have "foldr ?fun (?term r) zero = zero" 
    proof (induct r, simp)
      case (Cons d r)
      have "foldr ?fun (?term (d # r)) zero = foldr ?fun ( (d,zero) # ?term r) zero" by (simp only: map_replicate_trivial, simp)
      also have "\<dots> = zero" using Cons by simp
      finally show ?case .
    qed
    thus "zero = foldr ?fun (zip r (map (\<lambda> k. zero) [0..<nr])) zero" by (simp add: nr)
  qed
next
  case (Cons v m)
  from this obtain ncc where nc: "nc = Suc ncc" and wf: "wf_mat nr ncc m" unfolding wf_mat_def by (auto simp: wf_vec_def)
  from nc `wf_vec nc c` obtain a cc where c: "c = a # cc" and wfc: "wf_vec ncc cc" unfolding wf_vec_def by (cases c, auto)
  have rec: "scalar_prod (map (\<lambda> k. scalar_prod r (m ! k)) [0..<ncc]) cc = scalar_prod r (map (\<lambda> k. scalar_prod (row m k) cc) [0..<nr])"
    by (rule Cons, rule wf, rule wfc)
  have id: "map (\<lambda>k. scalar_prod r ((v # m) ! k)) [0..<Suc ncc] = scalar_prod r v # map (\<lambda> k. scalar_prod r (m ! k)) [0..<ncc]" by (induct ncc, auto)    
  from wfr have nr: "nr = length r" unfolding wf_vec_def by auto
  with Cons have v: "length v = length r" unfolding wf_mat_def by (auto simp: wf_vec_def)
  have "\<forall> i < nr. wf_vec ncc (row m i)" by (intro allI impI, rule wf_row[OF wf], simp)
  obtain tm where tm: "tm = transpose nr m" by auto
  hence idk: "\<forall> k < length r. row m k = tm ! k" using col_transpose_is_row[OF wf] unfolding col_def by (auto simp: nr)
  hence idtm1: "map (\<lambda>k. scalar_prod (row m k) cc) [0..<length r] = map (\<lambda>k. scalar_prod (tm ! k) cc) [0..<length r]" 
    and idtm2: "map (\<lambda>k. plus (times (v ! k) a) (scalar_prod (row m k) cc)) [0..<length r] = map (\<lambda>k. plus (times (v ! k) a) (scalar_prod (tm ! k) cc)) [0..<length r]" by auto
  from tm wf_transpose[OF wf] have "wf_mat ncc nr tm" by simp
  with nr have "length tm = length r" and  "(\<forall> i < length r. length (tm ! i) = ncc)" unfolding wf_mat_def by (auto simp: wf_vec_def) 
  with v have main: "plus (times (scalar_prod r v) a) (scalar_prod r (map (\<lambda>k. scalar_prod (tm ! k) cc) [0..<length r])) =
    scalar_prod r (map (\<lambda>k. plus (times (v ! k) a) (scalar_prod (tm ! k) cc)) [0..<length r])" 
  proof (induct r arbitrary: v tm)
    case Nil
    thus ?case by (auto simp: scalar_prodI_def row_def)
  next
    case (Cons b r)
    from this obtain c vv where v: "v = c # vv" and vvlen: "length vv = length r" by (cases v, auto)
    from Cons obtain u mm where tm: "tm = u # mm" and mmlen: "length mm = length r"  by (cases tm, auto)
    from Cons tm have argLen: "\<forall> i < length r. length (mm ! i) = ncc" by auto
    have rec: "plus (times (scalar_prod r vv) a) (scalar_prod r (map (\<lambda>k. scalar_prod (mm ! k) cc) [0..<length r])) =
     scalar_prod r (map (\<lambda>k. plus (times (vv ! k) a) (scalar_prod (mm ! k) cc)) [0..<length r])" 
      (is "plus (times ?rv a) ?recl = ?recr")
      by (rule Cons, auto simp: vvlen mmlen argLen)
    have id: "map (\<lambda>k. scalar_prod ((u # mm) ! k) cc) [0..<length (b # r)] = scalar_prod u cc # map (\<lambda>k. scalar_prod (mm ! k) cc) [0..<length r]" 
      by (simp, induct r, auto)
    have id2: "map (\<lambda>k. plus (times ((c # vv) ! k) a) (scalar_prod ((u # mm) ! k) cc)) [0..<length (b # r)] = 
               (plus (times c a) (scalar_prod u cc)) #
               map (\<lambda>k. plus (times (vv ! k) a) (scalar_prod (mm ! k) cc)) [0..<length r]" 
      by (simp, induct r, auto)
    show ?case proof (simp only: v tm, simp only: id, simp only: id2, simp only: scalar_prod_cons)
      let ?uc = "scalar_prod u cc"
      let ?bca = "times (times b c) a"
      have "plus (times (plus (times b c) ?rv) a) (plus (times b ?uc) ?recl) = plus (plus ?bca (times ?rv a)) (plus (times b ?uc) ?recl)" 
	by (simp add: left_distrib)
      also have "\<dots> = plus (plus ?bca (times ?rv a)) (plus ?recl (times b ?uc))" by (simp add: add_commute)
      also have "\<dots> = plus ?bca (plus (plus (times ?rv a) ?recl) (times b ?uc))" by (simp add: add_assoc)
      also have "\<dots> = plus ?bca (plus ?recr (times b ?uc))" by (simp only: rec)
      also have "\<dots> = plus ?bca (plus (times b ?uc) ?recr)" by (simp add: add_commute)
      also have "\<dots> = plus (times b (plus (times c a) ?uc)) ?recr" by (simp add: right_distrib mult_assoc add_assoc)
      finally show "plus (times (plus (times b c) ?rv) a) (plus (times b ?uc) ?recl) = plus (times b (plus (times c a) ?uc)) ?recr" .
    qed
  qed
  show ?case 
    by (simp add: c scalar_prod_cons, simp only: nc, simp only: id, simp add: scalar_prod_cons, simp only: rec, simp only: nr, simp only: idtm1 idtm2, simp only: main, simp only: idtm2[symmetric], simp add: row_def scalar_prod_cons)
qed


lemma mat_mult_assoc: 
  assumes wf1: "wf_mat nr n1 m1"
  and wf2: "wf_mat n1 n2 m2"
  and wf3: "wf_mat n2 nc m3"
  shows "mat_mult nr (mat_mult nr m1 m2) m3 = mat_mult nr m1 (mat_mult n1 m2 m3)" (is "?m12_3 = ?m1_23")
proof -
  let ?m12 = "mat_mult nr m1 m2"
  let ?m23 = "mat_mult n1 m2 m3"
  from wf1 wf2 have wf12: "wf_mat nr n2 ?m12" by (rule wf_mat_mult)
  from wf2 wf3 have wf23: "wf_mat n1 nc ?m23" by (rule wf_mat_mult)
  from wf1 wf23 have wf1_23: "wf_mat nr nc ?m1_23" by (rule wf_mat_mult)
  from wf12 wf3 have wf12_3: "wf_mat nr nc ?m12_3" by (rule wf_mat_mult)
  show ?thesis
  proof (simp only: mat_col_eq[OF wf12_3 wf1_23, symmetric], unfold col_def, intro allI impI)
    fix i
    assume i: "i < nc"
    with wf1_23 wf12_3 wf3 have len: "length (?m12_3 ! i) = length (?m1_23 ! i)" and ilen: "i < length m3" unfolding wf_mat_def by (auto simp: wf_vec_def)
    show "?m12_3 ! i = ?m1_23 ! i"
    proof (rule nth_equalityI[OF len], intro allI impI)
      fix j
      assume jlen: "j < length (?m12_3 ! i)"
      with wf12_3 i have j: "j < nr" unfolding wf_mat_def by (auto simp: wf_vec_def)      
      show "?m12_3 ! i ! j = ?m1_23 ! i ! j"
	by (simp only: mat_mult_to_scalar[OF wf12 wf3 j i],
	     simp only: mat_mult_to_scalar[OF wf1 wf23 j i], 
	     simp only: row_mat_mult_to_scalar[OF wf1 wf2 j],
	     simp only: col_mat_mult_to_scalar[OF wf2 wf3 i], 
	     simp only: scalar_product_assoc[OF wf2 wf_row[OF wf1 j] wf_col[OF wf3 i]])
    qed
  qed
qed

lemma mat_mult_assoc_n:  
  assumes wf1: "wf_mat n n m1"
  and wf2: "wf_mat n n m2"
  and wf3: "wf_mat n n m3"
  shows "mat_mult n (mat_mult n m1 m2) m3 = mat_mult n m1 (mat_mult n m2 m3)"
using assms
 by (rule mat_mult_assoc)


lemma scalar_left_zero : "scalar_prod (vec0 nn) v = zero"
  unfolding vec0I_def scalar_prodI_def
proof (induct nn arbitrary: v, simp)
  case (Suc m)
  thus ?case by (cases v, auto)
qed

lemma scalar_right_zero : "scalar_prod v (vec0 nn) = zero"
  unfolding vec0I_def scalar_prodI_def
proof (induct v arbitrary: nn, simp)
  case (Cons a vv)
  thus ?case by (cases nn, auto)
qed

lemma mat0_mult_left: assumes wf: "wf_mat nc ncc m"
  shows "mat_mult nr (mat0 nr nc) m = (mat0 nr ncc)"
proof (simp only: mat_eq_index[OF wf_mat_mult[OF wf_mat0 wf] wf_mat0], intro allI impI)
  fix i j
  assume i: "i < ncc" and j: "j < nr"
  show "mat_mult nr (mat0 nr nc) m ! i ! j = mat0 nr ncc ! i ! j"
  by (simp only: mat_mult_to_scalar[OF wf_mat0 wf j i], 
         simp only: mat0_index[OF i j], 
         simp only: mat0_row[OF j],
         simp only: scalar_left_zero)
qed


lemma mat0_mult_right: assumes wf: "wf_mat nr nc m"
  shows "mat_mult nr m (mat0 nc ncc) = (mat0 nr ncc)"
proof (simp only: mat_eq_index[OF wf_mat_mult[OF wf wf_mat0] wf_mat0], intro allI impI)
  fix i j
  assume i: "i < ncc" and j: "j < nr"
  show "mat_mult nr m (mat0 nc ncc) ! i ! j = mat0 nr ncc ! i ! j"
    by (simp only: mat_mult_to_scalar[OF wf wf_mat0 j i],
         simp only: mat0_index[OF i j],
         simp only: mat0_col[OF i],
         simp only: scalar_right_zero)
qed

lemma scalar_vec_plus_distrib_right: 
  assumes wf1: "wf_vec nr u"
  assumes wf2: "wf_vec nr v"
  assumes wf3: "wf_vec nr w"
  shows "scalar_prod u (vec_plus v w) = plus (scalar_prod u v) (scalar_prod u w)"
using assms
unfolding wf_vec_def scalar_prodI_def vec_plusI_def
proof (induct nr arbitrary: u v w, simp)
  case (Suc n)
  from Suc obtain a uu where u: "u = a # uu" by (cases u, auto)
  from Suc obtain b vv where v: "v = b # vv" by (cases v, auto)
  from Suc obtain c ww where w: "w = c # ww" by (cases w, auto)
  from Suc u v w have lu: "length uu = n" and lv: "length vv = n" and lw: "length ww = n" by auto
  show ?case by (simp only: u v w, simp, simp only: Suc(1)[OF lu lv lw], simp add: add_commute[of _ "times a c"] right_distrib add_assoc[symmetric])
qed

lemma scalar_vec_plus_distrib_left: 
  assumes wf1: "wf_vec nr u"
  assumes wf2: "wf_vec nr v"
  assumes wf3: "wf_vec nr w"
  shows "scalar_prod (vec_plus u v) w = plus (scalar_prod u w) (scalar_prod v w)"
using assms
unfolding wf_vec_def scalar_prodI_def vec_plusI_def
proof (induct nr arbitrary: u v w, simp)
  case (Suc n)
  from Suc obtain a uu where u: "u = a # uu" by (cases u, auto)
  from Suc obtain b vv where v: "v = b # vv" by (cases v, auto)
  from Suc obtain c ww where w: "w = c # ww" by (cases w, auto)
  from Suc u v w have lu: "length uu = n" and lv: "length vv = n" and lw: "length ww = n" by auto
  show ?case by (simp only: u v w, simp, simp only: Suc(1)[OF lu lv lw], simp add: add_commute[of _ "times b c"] left_distrib add_assoc[symmetric])
qed

lemma mat_mult_plus_distrib_right: 
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nc ncc m2"
  and wf3: "wf_mat nc ncc m3"
  shows "mat_mult nr m1 (mat_plus m2 m3) = mat_plus (mat_mult nr m1 m2) (mat_mult nr m1 m3)" (is "mat_mult nr m1 ?m23 = mat_plus ?m12 ?m13")
proof -
  let ?m1_23 = "mat_mult nr m1 ?m23"
  let ?m12_13 = "mat_plus ?m12 ?m13"
  from wf_mat_plus[OF wf2 wf3] have wf23: "wf_mat nc ncc ?m23" .
  from wf_mat_mult[OF wf1 wf2] have wf12: "wf_mat nr ncc ?m12" .
  from wf_mat_mult[OF wf1 wf3] have wf13: "wf_mat nr ncc ?m13" .
  from wf_mat_mult[OF wf1 wf23] have wf1_23: "wf_mat nr ncc ?m1_23" .
  from wf_mat_plus[OF wf12 wf13] have wf12_13: "wf_mat nr ncc ?m12_13" .
  show ?thesis 
  proof (simp only: mat_eq_index[OF wf1_23 wf12_13], intro impI allI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    show "?m1_23 ! i ! j = ?m12_13 ! i ! j"
      by (simp only: mat_mult_to_scalar[OF wf1 wf23 j i],
           simp only: mat_plus_index[OF wf12 wf13 i j],
	   simp only: mat_mult_to_scalar[OF wf1 wf2 j i],
	   simp only: mat_mult_to_scalar[OF wf1 wf3 j i],
	   simp only: col_mat_plus[OF wf2 wf3 i],
	rule scalar_vec_plus_distrib_right[OF wf_row[OF wf1 j] wf_col[OF wf2 i] wf_col[OF wf3 i]])
  qed
qed

lemma mat_mult_plus_distrib_left: 
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  and wf3: "wf_mat nc ncc m3"
  shows "mat_mult nr (mat_plus m1 m2) m3 = mat_plus (mat_mult nr m1 m3) (mat_mult nr m2 m3)" (is "mat_mult nr ?m12 _ = mat_plus ?m13 ?m23")
proof -
  let ?m12_3 = "mat_mult nr ?m12 m3"
  let ?m13_23 = "mat_plus ?m13 ?m23"
  from wf_mat_plus[OF wf1 wf2] have wf12: "wf_mat nr nc ?m12" .
  from wf_mat_mult[OF wf1 wf3] have wf13: "wf_mat nr ncc ?m13" .
  from wf_mat_mult[OF wf2 wf3] have wf23: "wf_mat nr ncc ?m23" .
  from wf_mat_mult[OF wf12 wf3] have wf12_3: "wf_mat nr ncc ?m12_3" .
  from wf_mat_plus[OF wf13 wf23] have wf13_23: "wf_mat nr ncc ?m13_23" .
  show ?thesis 
  proof (simp only: mat_eq_index[OF wf12_3 wf13_23], intro impI allI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    show "?m12_3 ! i ! j = ?m13_23 ! i ! j"
      by (simp only: mat_mult_to_scalar[OF wf12 wf3 j i],
           simp only: mat_plus_index[OF wf13 wf23 i j],
	   simp only: mat_mult_to_scalar[OF wf1 wf3 j i],
	   simp only: mat_mult_to_scalar[OF wf2 wf3 j i],
	   simp only: row_mat_plus[OF wf1 wf2 j],
	   rule scalar_vec_plus_distrib_left[OF wf_row[OF wf1 j] wf_row[OF wf2 j] wf_col[OF wf3 i]])
  qed
qed
end

context semiring_1
begin
abbreviation vec1 :: "nat \<Rightarrow> nat \<Rightarrow> 'a vec"
where "vec1 \<equiv> vec1I zero one"

abbreviation mat1 :: "nat \<Rightarrow> 'a mat"
where "mat1 \<equiv> mat1I zero one"


lemma scalar_left_one : assumes wf: "wf_vec nn v"
  and i: "i < nn"
  shows "scalar_prod (vec1I zero one nn i) v = v ! i"
  using assms 
  unfolding vec1I_def wf_vec_def 
proof (induct nn arbitrary: v i, blast)
  case (Suc n) note oSuc = this
  from this obtain a vv where v: "v = a # vv" and lvv: "length vv = n" by (cases v, auto)
  show ?case 
  proof (cases i)
    case 0
    thus ?thesis using scalar_left_zero unfolding vec0I_def by (simp add: v scalar_prod_cons add_commute)
  next
    case (Suc ii)
    thus ?thesis using oSuc lvv v by (auto simp: scalar_prod_cons)
  qed
qed


lemma scalar_right_one : assumes wf: "wf_vec nn v"
  and i: "i < nn"
  shows "scalar_prod v (vec1I zero one nn i) = v ! i"
  using assms 
  unfolding vec1I_def wf_vec_def 
proof (induct nn arbitrary: v i, blast)
  case (Suc n) note oSuc = this
  from this obtain a vv where v: "v = a # vv" and lvv: "length vv = n" by (cases v, auto)
  show ?case 
  proof (cases i)
    case 0
    thus ?thesis using scalar_right_zero unfolding vec0I_def by (simp add: v scalar_prod_cons add_commute)
  next
    case (Suc ii)
    thus ?thesis using oSuc lvv v by (auto simp: scalar_prod_cons)
  qed
qed


lemma mat1_mult_right: assumes wf: "wf_mat nr nc m"
  shows "mat_mult nr m (mat1I zero one nc) = m"
proof (simp only: mat_eq_index[OF wf_mat_mult[OF wf wf_mat1] wf], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "mat_mult nr m (mat1I zero one nc) ! i ! j = m ! i ! j"
    by (simp only: mat_mult_to_scalar[OF wf wf_mat1 j i],
    simp only: col_mat1[OF i],
    simp only: scalar_right_one[OF wf_row[OF wf j] i],
    simp only: row_col[OF wf j i],
    unfold col_def, simp)
qed


lemma mat1_mult_left: assumes wf: "wf_mat nr nc m"
  shows "mat_mult nr (mat1I zero one nr) m = m"
proof (simp only: mat_eq_index[OF wf_mat_mult[OF wf_mat1 wf] wf], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "mat_mult nr (mat1I zero one nr) m ! i ! j = m ! i ! j"
    by (simp only: mat_mult_to_scalar[OF wf_mat1 wf j i],
      simp only: row_mat1[OF j],
      simp only: scalar_left_one[OF wf_col[OF wf i] j], unfold col_def, simp)
qed
end

(* preorders are not used, since the condition 
   (x > y = x \<ge> y \<and> \<not> y \<ge> x)
   is sometimes not applicable. E.g., for polynomials over the naturals we have 2x \<ge> x, \<not> x \<ge> 2x, but \<not> 2x > x for x = 0
*)
class non_strict_order = 
  fixes ge :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 21)
  assumes ge_refl: "x \<ge> (x :: 'a)"
  and ge_trans[trans]: "\<lbrakk>x \<ge> y; (y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x \<ge> z"
begin
abbreviation vec_ge :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_ge \<equiv> vec_geI ge"

abbreviation mat_ge :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_ge \<equiv> mat_geI ge"

      
lemma vec_ge_refl: "vec_ge v v"
  unfolding vec_geI_def 
  by (induct v, auto simp: ge_refl)

lemma mat_ge_refl: "mat_ge m m"
  unfolding mat_geI_def
  by (induct m, auto simp: vec_ge_refl)

lemma vec_ge_trans: assumes ge12: "vec_ge v1 v2" and ge23: "vec_ge v2 v3" and wf_1: "wf_vec nr v1" and wf_2: "wf_vec nr v2" and wf_3: "wf_vec nr v3"
  shows "vec_ge v1 v3"
proof (simp only: vec_ge_index[OF wf_1 wf_3], intro allI impI) 
  fix i
  assume i: "i < nr"
  show "v1 ! i \<ge> v3 ! i"
    using 
      spec[OF ge12[simplified vec_ge_index[OF wf_1 wf_2]], of i]
      spec[OF ge23[simplified vec_ge_index[OF wf_2 wf_3]], of i]
      ge_trans[of "v1 ! i" "v2 ! i" "v3 ! i"]
      i 
    by blast
qed

lemma mat_ge_trans: assumes ge12: "mat_ge v1 v2" and ge23: "mat_ge v2 v3" and wf_1: "wf_mat nr nc v1" and wf_2: "wf_mat nr nc v2" and wf_3: "wf_mat nr nc v3"
  shows "mat_ge v1 v3"
proof (simp only: mat_ge_index[OF wf_1 wf_3], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "v1 ! i ! j \<ge> v3 ! i ! j"
    using 
      spec[OF ge12[simplified mat_ge_index[OF wf_1 wf_2]], of i]
      spec[OF ge23[simplified mat_ge_index[OF wf_2 wf_3]], of i]
      ge_trans[of "v1 ! i ! j" "v2 ! i ! j" "v3 ! i ! j"]
      i j
    by blast
qed
end

  
class ordered_ab_semigroup = non_strict_order + ab_semigroup_add +
  assumes plus_left_mono: "\<lbrakk>ge (x :: 'a) y\<rbrakk> \<Longrightarrow> ge (plus x z)  (plus y z)"
begin

lemma plus_right_mono: "ge y z \<Longrightarrow> ge (x + y) (x + z)"
  by (simp add: add_commute[of x], rule plus_left_mono, auto)

lemma vec_plus_left_mono: assumes ge: "vec_ge v1 v2" and wf_1: "wf_vec nr v1" and wf_2: "wf_vec nr v2" and wf_3: "wf_vec nr v3"
  shows "vec_ge (vec_plus v1 v3) (vec_plus v2 v3)"
by (simp only: vec_ge_index[OF wf_vec_plus[OF wf_1 wf_3] wf_vec_plus[OF wf_2 wf_3]], intro allI impI,
    simp only: vec_plus_index[OF wf_1 wf_3],
    simp only: vec_plus_index[OF wf_2 wf_3],
    rule plus_left_mono,
    auto simp: ge[simplified vec_ge_index[OF wf_1 wf_2]])

lemma mat_plus_left_mono: assumes ge: "mat_ge m1 m2" and wf_1: "wf_mat nr nc m1" and wf_2: "wf_mat nr nc m2" and wf_3: "wf_mat nr nc m3" 
  shows "mat_ge (mat_plus m1 m3) (mat_plus m2 m3)"
by (simp only: mat_ge_index[OF wf_mat_plus[OF wf_1 wf_3] wf_mat_plus[OF wf_2 wf_3]], intro allI impI,
    simp only: mat_plus_index[OF wf_1 wf_3],
    simp only: mat_plus_index[OF wf_2 wf_3],
    rule plus_left_mono,
    auto simp: ge[simplified mat_ge_index[OF wf_1 wf_2]])
end

class ordered_semiring_0 = ordered_ab_semigroup + semiring_0 +
 assumes times_left_mono: "\<lbrakk>ge z 0; ge x y\<rbrakk> \<Longrightarrow> ge (x * z) (y * z)"
     and times_right_mono: "\<lbrakk>ge x 0; ge y z\<rbrakk> \<Longrightarrow> ge (x * y) (x * z)"
begin


lemma scalar_prod_mono_left: assumes wf1: "wf_vec nr v1"
  and wf2: "wf_vec nr v2"
  and wf3: "wf_vec nr v3"
  and ge1: "vec_ge v1 v2"
  and ge2: "vec_ge v3 (vec0 nr)"
  shows "ge (scalar_prod v1 v3) (scalar_prod v2 v3)"
using assms unfolding wf_vec_def vec_geI_def vec0I_def
proof (simp)
  assume "length v1 = nr" and "length v2 = nr" and " length v3 = nr" and " \<forall>x\<in>set (zip v1 v2). split ge x" and " \<forall>x\<in>set (zip v3 (replicate nr 0)). split ge x"
  thus "ge (scalar_prod v1 v3) (scalar_prod v2 v3)"
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def ge_refl)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc have rec: "ge (scalar_prod w1 w3) (scalar_prod w2 w3)" (is "ge ?l ?r")
      by (auto simp: w1 w2 w3 v1 v2 v3)
    show ?case proof (simp add: v1 v2 v3 scalar_prod_cons)
      have one: "ge (a1 * a3) (a2 * a3)" using times_left_mono[of a3 a1 a2] Suc v1 v2 v3 by auto
      hence "ge (a1 * a3 + ?l) (a2 * a3 + ?l)" by (rule plus_left_mono)
      also have "ge \<dots> (a2 * a3 + ?r)" using rec by (rule plus_right_mono)
      finally show "ge (a1 * a3 + ?l) (a2 * a3 + ?r)" .
    qed
  qed
qed

lemma scalar_prod_mono_right: assumes wf1: "wf_vec nr v1"
  and wf2: "wf_vec nr v2"
  and wf3: "wf_vec nr v3"
  and ge1: "vec_ge v2 v3"
  and ge2: "vec_ge v1 (vec0 nr)"
  shows "ge (scalar_prod v1 v2) (scalar_prod v1 v3)"
using assms unfolding wf_vec_def vec_geI_def vec0I_def
proof (simp)
  assume "length v1 = nr" and "length v2 = nr" and " length v3 = nr" and " \<forall>x\<in>set (zip v2 v3). split ge x" and " \<forall>x\<in>set (zip v1 (replicate nr 0)). split ge x"
  thus "ge (scalar_prod v1 v2) (scalar_prod v1 v3)"
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def ge_refl)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc have rec: "ge (scalar_prod w1 w2) (scalar_prod w1 w3)" (is "ge ?l ?r")
      by (auto simp: w1 w2 w3 v1 v2 v3)
    show ?case proof (simp add: v1 v2 v3 scalar_prod_cons)
      have one: "ge (a1 * a2)  (a1 * a3)" using times_right_mono[of a1 a2 a3] Suc v1 v2 v3 by auto
      hence "ge (a1 * a2 + ?l) (a1 * a3 + ?l)" by (rule plus_left_mono)
      also have "ge \<dots> (a1 * a3 + ?r)" using rec by (rule plus_right_mono)
      finally show "ge (a1 * a2 + ?l) (a1 * a3 + ?r)" .
    qed
  qed
qed


lemma mat_mult_left_mono:
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nr nc m2"
  and wf3: "wf_mat nc ncc m3"
  and ge1: "mat_ge m1 m2"
  and ge2: "mat_ge m3 (mat0 nc ncc)"
  shows "mat_ge (mat_mult nr m1 m3) (mat_mult nr m2 m3)"
proof -
  let ?m13 = "mat_mult nr m1 m3"
  let ?m23 = "mat_mult nr m2 m3"
  from wf_mat_mult[OF wf1 wf3] have wf13: "wf_mat nr ncc ?m13" .
  from wf_mat_mult[OF wf2 wf3] have wf23: "wf_mat nr ncc ?m23" .
  show ?thesis 
  proof (simp only: mat_ge_index[OF wf13 wf23], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    from ge1 have ge1a: "\<forall>i<nc. \<forall> j < nr.  ge (m1 ! i ! j) (m2 ! i ! j)"
      using mat_ge_index[OF wf1 wf2] by simp
    from ge2 have ge2a: "\<forall>ia<nc. ge (col m3 i ! ia)  (vec0 nc ! ia)"
      using mat_ge_index[OF wf3 wf_mat0] i unfolding col_def mat0I_def vec0I_def
      by auto      
    show "ge (?m13 ! i ! j)  (?m23 ! i ! j)"
      by (simp only: mat_mult_to_scalar[OF wf1 wf3 j i],
	simp only: mat_mult_to_scalar[OF wf2 wf3 j i], 
	rule scalar_prod_mono_left[OF wf_row[OF wf1 j] wf_row[OF wf2 j] wf_col[OF wf3 i]],
	simp only: vec_ge_index[OF wf_row[OF wf1 j] wf_row[OF wf2 j]],
	simp add: row_col[OF wf1 j] row_col[OF wf2 j] col_def, (auto simp: ge1a j)[1],
	simp only: vec_ge_index[OF wf_col[OF wf3 i] wf_vec0],
	rule ge2a)
  qed
qed  


lemma mat_mult_right_mono:
  assumes wf1: "wf_mat nr nc m1"
  and wf2: "wf_mat nc ncc m2"
  and wf3: "wf_mat nc ncc m3"
  and ge1: "mat_ge m1 (mat0 nr nc)"
  and ge2: "mat_ge m2 m3"
  shows "mat_ge (mat_mult nr m1 m2) (mat_mult nr m1 m3)"
proof -
  let ?m12 = "mat_mult nr m1 m2"
  let ?m13 = "mat_mult nr m1 m3"
  from wf_mat_mult[OF wf1 wf2] have wf12: "wf_mat nr ncc ?m12" .
  from wf_mat_mult[OF wf1 wf3] have wf13: "wf_mat nr ncc ?m13" .
  show ?thesis 
  proof (simp only: mat_ge_index[OF wf12 wf13], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    from ge2 i have ge2a: " \<forall>ia<nc. ge (col m2 i ! ia) (col m3 i ! ia)"
      using mat_ge_index[OF wf2 wf3] unfolding col_def by auto
    from ge1 j have ge1a: " \<forall>i<nc. ge (m1 ! i ! j) 0" 
      using mat_ge_index[OF wf1 wf_mat0] unfolding mat0I_def vec0I_def
      by auto
    show "ge (?m12 ! i ! j) (?m13 ! i ! j)"
      by  (simp only: mat_mult_to_scalar[OF wf1 wf2 j i],
	simp only: mat_mult_to_scalar[OF wf1 wf3 j i],
	rule scalar_prod_mono_right[OF wf_row[OF wf1 j] wf_col[OF wf2 i] wf_col[OF wf3 i]], 
	simp only: vec_ge_index[OF wf_col[OF wf2 i] wf_col[OF wf3 i]], rule ge2a, 
	simp only: vec_ge_index[OF wf_row[OF wf1 j] wf_vec0],
	simp add: row_col[OF wf1 j] vec0I_def col_def, rule ge1a) 
  qed
qed
end

class ordered_semiring_1 = ordered_semiring_0 + semiring_1 +
  assumes one_ge_zero: "ge 1 0"
begin

lemma vec1_ge_vec0: assumes "i < n" shows "vec_ge (vec1 n i) (vec0 n)"
unfolding vec_geI_def vec0I_def vec1I_def using assms
proof (auto simp: set_zip)
  fix j
  assume i: "i < n" and j: "j < n"
  show "ge ((replicate i 0 @ 1 # replicate (n - Suc i) 0) ! j) 0" (is "ge (?repl j) _")
  proof (cases "j < i", simp add: nth_append ge_refl)
    case False
    show ?thesis
    proof (cases "j - i", simp add: nth_append ge_refl one_ge_zero)
      case (Suc k)
      with False
      have repl: "?repl j = (replicate (n - Suc i) 0) ! k" by (simp add: nth_append)
      from Suc False j have "k < (n - Suc i)" by arith
      with repl show ?thesis by (auto simp: one_ge_zero ge_refl)
    qed
  qed
qed

lemma mat1_ge_mat0: shows "mat_ge (mat1 n) (mat0 n n)"
unfolding mat_geI_def mat0I_def mat1I_def
by (auto simp: set_zip vec1_ge_vec0)
end

class max_ordered_monoid_add = ordered_ab_semigroup + ab_semigroup_add + monoid_add + 
  fixes max0 :: "'a \<Rightarrow> 'a"
  assumes max0_id_pos: "ge x 0 \<Longrightarrow> max0 x = x"
  and max0_pos: " ge (max0 x) 0"
  and max0_mono: "ge x y \<Longrightarrow> ge (max0 x) (max0 y)"
  and max0_x: "ge (max0 x) x"
begin

abbreviation vec_max0 :: "'a vec \<Rightarrow> 'a vec"
where "vec_max0 \<equiv> vec_max0I max0"

abbreviation mat_max0 :: "'a mat \<Rightarrow> 'a mat"
where "mat_max0 \<equiv> mat_max0I max0"


lemma vec_max0_x: "vec_ge (vec_max0 v) v"
unfolding vec_geI_def vec_max0I_def
by (auto simp: set_zip max0_x)

lemma mat_max0_x: "mat_ge (mat_max0 m) m"
unfolding mat_geI_def mat_max0I_def
by (auto simp: set_zip vec_max0_x)

lemma vec_max0_pos: assumes "wf_vec nr v"
  shows "vec_ge (vec_max0 v) (vec0 nr)"
using assms unfolding wf_vec_def vec0I_def vec_max0I_def vec_geI_def
by (auto simp: set_zip max0_pos)

lemma mat_max0_pos: assumes "wf_mat nr nc v"
  shows "mat_ge (mat_max0 v) (mat0 nr nc)"
using assms unfolding wf_mat_def mat0I_def mat_max0I_def mat_geI_def
by (auto simp: set_zip vec_max0_pos)

lemma vec_max0_id_pos: assumes "vec_ge v (vec0 nr)" and "wf_vec nr v"  
  shows "vec_max0 v = v"
using assms unfolding wf_vec_def vec0I_def vec_max0I_def vec_geI_def 
by (induct v arbitrary: nr, auto simp: max0_id_pos)

lemma mat_max0_id_pos: assumes "mat_ge m (mat0 nr nc)" and "wf_mat nr nc m"  
  shows "mat_max0 m = m"
using assms unfolding wf_mat_def mat0I_def mat_max0I_def mat_geI_def 
by (induct m arbitrary: nc, auto simp: vec_max0_id_pos)

lemma vec_max0_mono: assumes "vec_ge v w" and "wf_vec nr v" and "wf_vec nr w"
  shows "vec_ge (vec_max0 v) (vec_max0 w)"
using assms unfolding wf_vec_def vec_max0I_def vec_geI_def
proof (simp, induct nr arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1) note oCons = this
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      with oCons Suc have len: "ge a b \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> ge a b)" by auto
      show ?thesis
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule max0_mono, simp add: len, rule Suc, auto simp: len)
    qed
  qed
qed

lemma mat_max0_mono: assumes "mat_ge v w" and "wf_mat nr nc v" and "wf_mat nr nc w"
  shows "mat_ge (mat_max0 v) (mat_max0 w)"
using assms unfolding wf_mat_def mat_max0I_def mat_geI_def
proof (simp, induct nc arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1) note oCons = this
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      with oCons Suc have len: "(vec_ge a b) \<and> wf_vec nr a \<and> wf_vec nr b \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a \<in> set v1. wf_vec nr a) \<and> (\<forall> b \<in> set w1. wf_vec nr b) \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> (vec_ge a b))" by auto
      show ?thesis 
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule vec_max0_mono[where nr = nr], (simp add: len)+, rule Suc, auto simp: len)
    qed
  qed
qed
end


(*
   we do not use a class here, but only a locale since we often 
   have parametric orders, e.g. on rational numbers there are several orders 
   > where x > y = x \<ge> y + \<delta> for some parameter \<delta>
*)
locale order_pair = 
  fixes gt :: "'a :: non_strict_order \<Rightarrow> 'a \<Rightarrow> bool" 
  assumes compat[trans]: "\<lbrakk>x \<ge> y; gt y z\<rbrakk> \<Longrightarrow> gt x z"
  and compat2[trans]: "\<lbrakk>gt x y; y \<ge> z\<rbrakk> \<Longrightarrow> gt x z"
begin

abbreviation mat_pre_gt :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_pre_gt \<equiv> mat_pre_gtI gt"

abbreviation mat_gt :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_gt \<equiv> mat_gtI ge gt"

lemma mat_gt_compat: assumes sd_n: "sd \<le> n" and  ge: "mat_ge m1 m2" and gt: "mat_gt sd m2 m3" and wf1: "wf_mat n n m1" and wf2: "wf_mat n n m2" and wf3: "wf_mat n n m3" 
  shows "mat_gt sd m1 m3"
proof -
  from gt[simplified mat_gt_index[OF wf2 wf3 sd_n]] obtain i j 
    where i: "i < sd" and j: "j < sd" and gt: "gt (m2 ! i ! j) (m3 ! i ! j)" and ge23: "mat_ge m2 m3" by auto 
  from ge[simplified mat_ge_index[OF wf1 wf2]] i j sd_n have geij: "m1 ! i ! j \<ge> m2 ! i ! j" by auto
  from mat_ge_trans[OF ge ge23 wf1 wf2 wf3] have ge: "mat_ge m1 m3" .
  with compat[OF geij gt] i j show ?thesis 
    by (simp only: mat_gt_index[OF wf1 wf3 sd_n], auto)
qed

lemma mat_gt_compat2: assumes sd_n: "sd \<le> n" and gt: "mat_gt sd m1 m2" and ge: "mat_ge m2 m3" and wf1: "wf_mat n n m1" and wf2: "wf_mat n n m2" and wf3: "wf_mat n n m3" 
  shows "mat_gt sd m1 m3"
proof -
  from gt[simplified mat_gt_index[OF wf1 wf2 sd_n]] obtain i j 
    where i: "i < sd" and j: "j < sd" and gt: "gt (m1 ! i ! j) (m2 ! i ! j)" and ge12: "mat_ge m1 m2" by auto 
  from ge[simplified mat_ge_index[OF wf2 wf3]] i j sd_n have geij: "m2 ! i ! j \<ge> m3 ! i ! j" by auto
  from mat_ge_trans[OF ge12 ge wf1 wf2 wf3] have ge: "mat_ge m1 m3" .
  with compat2[OF gt geij] i j show ?thesis 
    by (simp only: mat_gt_index[OF wf1 wf3 sd_n], auto)
qed
end

locale one_mono_ordered_semiring_1 = order_pair gt 
  for gt :: "'a :: ordered_semiring_1 \<Rightarrow> 'a \<Rightarrow> bool" + 
  assumes plus_gt_left_mono: "gt x y \<Longrightarrow> gt (x + z) (y + z)"
  and one_gt_zero: "gt 1 0"
begin 

lemma mat_plus_gt_left_mono: assumes sd_n: "sd \<le> n" and gt: "mat_gt sd m1 m2" and ge: "mat_ge m3 m4" and wf1: "wf_mat n n m1" and wf2: "wf_mat n n m2" and wf3: "wf_mat n n m3" and wf4: "wf_mat n n m4"
  shows "mat_gt sd (mat_plus m1 m3) (mat_plus m2 m4)"
proof -
  let ?m13 = "mat_plus m1 m3"
  let ?m23 = "mat_plus m2 m3"
  let ?m32 = "mat_plus m3 m2"
  let ?m24 = "mat_plus m2 m4"
  let ?m42 = "mat_plus m4 m2"
  have wf13: "wf_mat n n ?m13" and wf24: "wf_mat n n ?m24" by ((rule wf_mat_plus[of n n], auto simp: wf1 wf2 wf3 wf4)+)
  from gt[simplified mat_gt_index[OF wf1 wf2 sd_n]] obtain i j where
    i: "i < sd" and j: "j < sd" and gt: "gt (m1 ! i ! j) (m2 ! i ! j)" and ge12: "mat_ge m1 m2" by auto
  with sd_n have ni: "i < n" and nj: "j < n" by auto
  from mat_plus_left_mono[OF ge12 wf1 wf2 wf3] have one: "mat_ge ?m13 ?m23" .
  from mat_plus_left_mono[OF ge wf3 wf4 wf2] have "mat_ge ?m32 ?m42" .
  hence two: "mat_ge ?m23 ?m24" by (simp add: mat_plus_comm[of m2 m3] mat_plus_comm[of m2 m4])
  have matge: "mat_ge ?m13 ?m24" by (rule mat_ge_trans[OF one two, of n n], (rule wf_mat_plus[of n n], auto simp: wf1 wf2 wf3 wf4)+)
  from i j sd_n ge[simplified mat_ge_index[OF wf3 wf4]] have ge: "m3 ! i ! j \<ge> m4 ! i ! j" by auto
  from compat2[OF plus_gt_left_mono[OF gt] plus_right_mono[OF ge]] mat_plus_index[OF wf1 wf3 ni nj] mat_plus_index[OF wf2 wf4 ni nj]      
  have gt: "gt (?m13 ! i ! j) (?m24 ! i ! j)" by simp
  from i j matge gt  show ?thesis 
    by (auto simp: mat_gt_index[OF wf13 wf24 sd_n] matge)
qed

lemma mat_plus_gt_ge_mono: assumes sd_n: "sd \<le> n" and  gt: "mat_gt sd m1 m2" and ge: "mat_ge m3 m4" and wf1: "wf_mat n n m1" and wf2: "wf_mat n n m2" and wf3: "wf_mat n n m3" and wf4: "wf_mat n n m4"
  shows "mat_gt sd (mat_plus m1 m3) (mat_plus m2 m4)"
  by (rule mat_gt_compat2[OF sd_n], rule mat_plus_gt_left_mono[OF sd_n gt ge], simp_all add: wf1 wf2 wf3 wf4 wf_mat_plus, rule mat_ge_refl) 

lemma mat1_gt_mat0: assumes sd_pos: "sd > 0" and sd_n: "sd \<le> n"
  shows "mat_gt sd (mat1 n) (mat0 n n)"
proof -
  from assms have n: "n > 0" by auto
  show ?thesis 
    by (simp add: mat_gt_index[OF wf_mat1 wf_mat0 sd_n] mat1_ge_mat0,
      (rule exI[of _ 0], simp add: sd_pos)+, simp add: mat1_index[OF n n] mat0_index[OF n n] one_gt_zero)
qed
end
    


locale SN_one_mono_ordered_semiring_1 = one_mono_ordered_semiring_1 + order_pair + 
  assumes SN: "SN {(x,y) . ((y :: 'a) \<ge> 0) \<and> (gt x y)}"
begin

abbreviation mat_ns :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succeq>m _ _)" [51,51,51] 50) 
 where "m1 \<succeq>m n m2 \<equiv> (wf_mat n n m1 \<and> wf_mat n n m2 \<and> mat_ge m1 m2)"

abbreviation mat_s :: "'a mat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succ>m _ _ _)" [51,51,51,51] 50)
 where "m1 \<succ>m n sd m2 \<equiv> (wf_mat n n m1 \<and> wf_mat n n m2 \<and> mat_ge m2 (mat0 n n) \<and> mat_gt sd m1 m2)"

lemma mat_sn : assumes sd_n: "sd \<le> n" shows "SN {(m1,m2) . m1 \<succ>m n sd m2}"
unfolding SN_defs 
proof clarify
  fix e f 
  assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(m1,m2). m1 \<succ>m n sd m2}"
  hence len_n: "\<And> i. length (f i) = n" by (auto simp: wf_mat_def) 
  let ?rel = "{(x,y). (y \<ge> 0) \<and> (gt x y)}"
  let ?gt = "\<lambda> k i j. gt (f k ! i ! j) (f (Suc k) ! i ! j)"
  let ?ge = "\<lambda> k i j. f k ! i ! j \<ge> f (Suc k) ! i ! j"
  let ?gez = "\<lambda> k i j. f (Suc k) ! i ! j \<ge> 0"
  let ?f = "\<lambda> k ij. f k ! (ij div sd) ! (ij mod sd)"
  let ?ij = "\<lambda> i j. (i * sd + j)"
  let ?fgt = "\<lambda> k ij. (?f k ij,?f (Suc k) ij) \<in> ?rel"
  let ?fpgt = "\<lambda> k ij. gt (?f k ij) (?f (Suc k) ij)"
  let ?fge = "\<lambda> k ij. ?f k ij \<ge> ?f (Suc k) ij"
  let ?sd = "sd * sd"
  have all: "\<And> k. (\<exists> ij < ?sd. ?fgt k ij) \<and> (\<forall> ij < ?sd. ?fge k ij)"
  proof -
    fix k
    from spec[OF ass, of k] have wf: "wf_mat n n (f k)" and wf1: "wf_mat n n (f (Suc k))" and gt: "mat_gt sd (f k) (f (Suc k))" and
      gez: "mat_ge (f (Suc k)) (mat0 n n)" by auto
    from gt[simplified mat_gt_index[OF wf wf1 sd_n]] obtain i j where i: "i < sd" and j: "j < sd" and ge: "mat_ge (f k) (f (Suc k))"
      and gt: "?gt k i j" by auto
    hence pgt: "?fpgt k (?ij i j)" by (auto simp: mul_div_eq[OF j] j)
    have ij: "?ij i j < ?sd" using i j smaller_product by auto
    from ge[simplified mat_ge_index[OF wf wf1]] sd_n have "\<forall> i < sd. \<forall> j < sd. ?ge k i j" by auto
    hence ge: "\<forall> ij < ?sd. ?fge k ij" by (simp add: all_all_into_all)      
    from gez[simplified mat_ge_index[OF wf1 wf_mat0]] sd_n have "\<forall> i < sd. \<forall> j < sd. ?gez k i j" by (auto simp: mat0_index)
    hence "?gez k i j" using i j by simp
    hence "?f (Suc k) (?ij i j) \<ge> 0" by (auto simp: mul_div_eq[OF j] j)
    with pgt have gt: "?fgt k (?ij i j)" by auto 
    from gt ge ij show "(\<exists> ij < ?sd. ?fgt k ij) \<and> (\<forall> ij < ?sd. ?fge k ij)" by auto
  qed
  obtain f sd where f: "f = ?f" and "sd = ?sd" by auto
  with all have ex: "\<And> k. (\<exists> i < sd. (f k i, f (Suc k) i) \<in> ?rel)" and all: "\<And> k.(\<forall> i < sd. f k i \<ge> f (Suc k) i)" by auto
  let ?g = "\<lambda> k i. (f k i, f (Suc k) i) \<in> ?rel"
  from ex have g: "\<forall> k. \<exists> i < sd. ?g k i" by auto
  from inf_drawer_principle[OF g] obtain i where i: "i < sd" and inf: "\<forall> k. \<exists> k' \<ge> k. ?g k' i" by auto
  let ?h = "\<lambda> k. (f k i)"
  let ?nRel = "{(x,y) | x y :: 'a. x \<ge> y}"
  from all i have all: "\<forall> k. (?h k, ?h (Suc k)) \<in> ?nRel \<union> ?rel" by auto
  from SN have SNe: "SN_elt ?rel (?h 0)" unfolding SN_defs by auto
  have comp: "?nRel O ?rel \<subseteq> ?rel" using compat by auto
  from non_strict_ending[OF all comp] SNe
  obtain j where "\<forall> k \<ge> j. (?h k, ?h (Suc k)) \<in> ?nRel - ?rel" by auto
  with inf  show False by blast
qed
end

locale SN_strict_mono_ordered_semiring_1 = SN_one_mono_ordered_semiring_1 +
  fixes mono :: "'a \<Rightarrow> bool"
  assumes mono: "\<lbrakk>mono x; gt y z; x \<ge> 0\<rbrakk> \<Longrightarrow> gt (x * y) (x * z)" 
  and one_mono: "mono 1"
begin

abbreviation mat_mono :: "nat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_mono \<equiv> mat_monoI mono"

lemma mat_mono_mat1: assumes sd_n: "sd \<le> n" shows "mat_mono sd (mat1 n)"
proof (simp add: mat_mono_index[OF wf_mat1 sd_n], intro allI impI)
  fix i
  assume i: "i < sd"
  with sd_n have n: "i < n" by auto
  have "mat1 n ! i ! i = (1 :: 'a)" using mat1_index[OF n n] by simp
  with i one_mono show "\<exists> j < sd. mono (mat1 n ! i ! j)" by auto
qed

lemma mat_mono: assumes sd_n: "sd \<le> n" and wf1: "wf_mat n n m1" and wf2: "wf_mat n n m2" and wf3: "wf_mat n n m3" and gt: "mat_gt sd m2 m3" and ge: "mat_ge m1 (mat0 n n)" and mmono: "mat_mono sd m1"
  shows "mat_gt sd (mat_mult n m1 m2) (mat_mult n m1 m3)"
proof -
  let ?m12 = "mat_mult n m1 m2"
  let ?m13 = "mat_mult n m1 m3"
  from wf1 wf2 have wf12: "wf_mat n n ?m12" by (rule wf_mat_mult)
  from wf1 wf3 have wf13: "wf_mat n n ?m13" by (rule wf_mat_mult)
  from mat_mult_right_mono[OF wf1 wf2 wf3 ge] gt have gem: "mat_ge ?m12 ?m13" unfolding mat_gtI_def by auto
  from gt obtain i j where i: "i < sd" and j: "j < sd" and gt: "gt (m2 ! i ! j) (m3 ! i ! j)" and geq: "mat_ge m2 m3"
    by (auto simp: mat_gt_index[OF wf2 wf3 sd_n])
  from mmono mat_mono_index[OF wf1 sd_n] j obtain k where k: "k < sd" and monojk: "mono (m1 ! j ! k)" by auto
  from i j k sd_n have ni: "i < n" and nj: "j < n" and nk: "k < n" by auto
  show ?thesis 
  proof (simp only: mat_gt_index[OF wf12 wf13 sd_n], simp add: gem, 
      rule exI[of _ i], simp add: i, 
      rule exI[of _ k], simp add: k)
    let ?r = "\<lambda> n. map (\<lambda> j. m1 ! j ! k) [0 ..< n]"
    from wf_row[OF wf1 nk] have wfr: "length (row m1 k) = n" unfolding wf_vec_def by auto
    from wf_col[OF wf2 ni] have wfc2: "length (col m2 i) = n" unfolding wf_vec_def by auto
    from wf_col[OF wf3 ni] have wfc3: "length (col m3 i) = n" unfolding wf_vec_def by auto
    have "row m1 k = map (\<lambda> i. row m1 k ! i) [0 ..< length (row m1 k)]" by (rule map_nth[symmetric])
    also have "\<dots> = map (\<lambda> i. row m1 k !  i) [0 ..< n]" by (simp add: wfr)
    also have "\<dots> = ?r n" using row_col[OF wf1 nk, simplified col_def]
      by auto
    finally have r: "row m1 k = ?r n" .
    let ?c2 = "\<lambda> n. map (\<lambda> j. m2 ! i ! j) [0 ..< n]"
    have c2: "col m2 i = ?c2 n" by (simp only: wfc2[symmetric] col_def, simp add: map_nth)
    let ?c3 = "\<lambda> n. map (\<lambda> j. m3 ! i ! j) [0 ..< n]"
    have c3: "col m3 i = ?c3 n" by (simp only: wfc3[symmetric] col_def, simp add: map_nth)
    from mat_mult_to_scalar[OF wf1 wf2 nk ni]
    have s12: "?m12 ! i ! k = scalar_prod (?r n) (?c2 n)" by (simp add: r c2)
    from mat_mult_to_scalar[OF wf1 wf3 nk ni]
    have s13: "?m13 ! i ! k = scalar_prod (?r n) (?c3 n)" by (simp add: r c3)
    have r0: "\<forall> j < n. ?r n ! j \<ge> 0" 
    proof (intro impI allI)
      fix j
      assume "j < n"
      with ge[simplified mat_ge_index[OF wf1 wf_mat0]] nk
      show "?r n ! j \<ge> 0" by (simp add: mat0_index)
    qed
    have c2c3: "\<forall> j < n. ?c2 n ! j \<ge> ?c3 n ! j"
    proof (intro impI allI)
      fix j
      assume "j < n"
      with ni geq[simplified mat_ge_index[OF wf2 wf3]] 
      show "?c2 n ! j \<ge> ?c3 n ! j" by simp
    qed
    from nj r0 c2c3 have "gt (scalar_prod (?r n) (?c2 n)) (scalar_prod (?r n) (?c3 n))"
    proof (induct n, simp)
      case (Suc n)
      have "scalar_prod (?r (Suc n)) (?c2 (Suc n)) = scalar_prod (?r n @ [m1 ! n ! k]) (?c2 n @ [m2 ! i ! n])" 
        (is "?sum2 = _") by simp
      also have "\<dots> = m1 ! n ! k * m2 ! i ! n + scalar_prod (?r n) (?c2 n)" (is "_ = plus ?p2 ?s2")
        by (simp add: scalar_prod_last)
      finally have sum2: "?sum2 = ?p2 + ?s2" .
      have "scalar_prod (?r (Suc n)) (?c3 (Suc n)) = scalar_prod (?r n @ [m1 ! n ! k]) (?c3 n @ [m3 ! i ! n])" 
        (is "?sum3 = _") by simp
      also have "\<dots> = m1 ! n ! k * m3 ! i ! n + scalar_prod (?r n) (?c3 n)" (is "_ = plus ?p3 ?s3") 
        by (simp add: scalar_prod_last)
      finally have sum3: "?sum3 = ?p3 + ?s3" .
      from Suc(3) have z: "m1 ! n ! k \<ge> 0" by (simp del: upt_Suc)
      from Suc(3) have za: "\<forall> j < n. ?r n ! j \<ge> 0"  by (simp del: upt_Suc)
      from Suc(4) have ge: "\<forall> j < n. ?c2 n ! j \<ge> ?c3 n ! j"  by (simp del: upt_Suc)
      show ?case
      proof (cases "j = n")
        case False
        with Suc(2) have j: "j < n" by auto
        have rec: "gt ?s2 ?s3"
          by (rule Suc, rule j, rule za, rule ge)
        from Suc(4) have ge: "m2 ! i ! n \<ge> m3 ! i ! n" by (simp del: upt_Suc)
        from times_right_mono[OF z ge] have p23: "?p2 \<ge> ?p3" .
        from compat2[OF plus_gt_left_mono[OF rec] plus_right_mono[OF p23]] have "gt (?s2 + ?p2) (?s3 + ?p3)" .
        with add_commute[of ?p2] add_commute[of ?p3] have "gt (?p2 + ?s2) (?p3 + ?s3)" by simp
        with sum2 sum3 show ?thesis by simp 
      next
        case True        
        with mono[OF monojk gt] z have p23: "gt ?p2 ?p3" by simp
        have wf1: "wf_vec n (?r n)" by (simp add: wf_vec_def)
        have wf2: "wf_vec n (?c2 n)" by (simp add: wf_vec_def)
        have wf3: "wf_vec n (?c3 n)" by (simp add: wf_vec_def)
        from ge have ge: "vec_ge (?c2 n) (?c3 n)" 
          by (simp only: vec_ge_index[OF wf2 wf3])
        from za have z: "vec_ge (?r n) (vec0 n)" 
          by (simp only: vec_ge_index[OF wf1 wf_vec0], auto simp: vec0I_def)
        have s23: "?s2 \<ge> ?s3"
          by (rule scalar_prod_mono_right, (simp add: wf_vec_def)+, rule ge, rule z)
        from compat2[OF plus_gt_left_mono[OF p23] plus_right_mono[OF s23]] sum2 sum3 show ?thesis by simp
      qed
    qed
    with s12 s13 show "gt (?m12 ! i ! k) (?m13 ! i ! k)" by simp
  qed    
qed
end

locale both_mono_ordered_semiring_1 = order_pair gt 
  for gt :: "'a :: ordered_semiring_1 \<Rightarrow> 'a \<Rightarrow> bool" + 
  assumes plus_gt_both_mono: "\<lbrakk>gt x y; gt z u\<rbrakk> \<Longrightarrow> gt (x + z) (y + u)"
  and times_gt_left_mono: "gt x y \<Longrightarrow> gt (x * z) (y * z)" 
  and zero_leastI: "gt x 0" 
  and zero_leastII: "gt 0 x \<Longrightarrow> x = 0" 
  and zero_leastIII: "(x :: 'a) \<ge> 0"
begin 

abbreviation mat_gt_arc :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_gt_arc \<equiv> mat_geI gt"

lemma scalar_prod_left_mono: assumes wf1: "wf_vec nr v1"
  and wf2: "wf_vec nr v2"
  and wf3: "wf_vec nr v3"
  and gt1: "vec_geI gt v1 v2"
  shows "gt (scalar_prod v1 v3)  (scalar_prod v2 v3)"
proof -
  from vec_ge_index[OF wf1 wf2] gt1 have g1: "\<forall> i < nr. gt (v1 ! i) (v2 ! i)" by auto
  from g1 wf1 wf2 wf3 show ?thesis unfolding wf_vec_def
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def zero_leastI)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc v1 v2 have a12: "gt a1 a2" and w12: "\<forall> i < nrr. gt (w1 ! i) (w2 ! i)" by auto
    have rec: "gt (scalar_prod w1 w3) (scalar_prod w2 w3)" 
      by (rule Suc, auto simp: w1 w2 w3 w12)
    have a: "gt (a1 * a3) (a2 * a3)" by (rule times_gt_left_mono[OF a12])
    show ?case 
      by (simp add: v1 v2 v3 scalar_prod_cons, rule plus_gt_both_mono[OF a rec])
  qed
qed


lemma mat_gt_arc_compat: assumes ge: "mat_ge m1 m2" and gt: "mat_gt_arc m2 m3" and wf1: "wf_mat nr nc m1" and wf2: "wf_mat nr nc m2" and wf3: "wf_mat nr nc m3" 
  shows "mat_gt_arc m1 m3"
proof (simp only: mat_ge_index[OF wf1 wf3], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  from ge i j have one: "m1 ! i ! j \<ge> m2 ! i ! j" using mat_ge_index[OF wf1 wf2] by auto
  from gt i j have two: "gt (m2 ! i ! j) (m3 ! i ! j)" using mat_ge_index[OF wf2 wf3] by auto
  from one two show "gt (m1 ! i ! j)  (m3 ! i ! j)" by (rule compat)
qed

lemma mat_gt_arc_compat2: assumes gt: "mat_gt_arc m1 m2" and ge: "mat_ge m2 m3" and wf1: "wf_mat nr nc m1" and wf2: "wf_mat nr nc m2" and wf3: "wf_mat nr nc m3" 
  shows "mat_gt_arc m1 m3"
proof (simp only: mat_ge_index[OF wf1 wf3], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  from gt i j have one: "gt (m1 ! i ! j) (m2 ! i ! j)" using mat_ge_index[OF wf1 wf2] by auto
  from ge i j have two: "m2 ! i ! j \<ge> m3 ! i ! j" using mat_ge_index[OF wf2 wf3] by auto
  from one two show "gt (m1 ! i ! j) (m3 ! i ! j)" by (rule compat2)
qed

lemma mat_gt_arc_plus_mono: assumes gt1: "mat_gt_arc x y"
  and gt2: "mat_gt_arc z u"
  and wfx: "wf_mat nr nc x"
  and wfy: "wf_mat nr nc y"
  and wfz: "wf_mat nr nc z"
  and wfu: "wf_mat nr nc u"
  shows "mat_gt_arc (mat_plus x z) (mat_plus y u)"
proof -
  let ?xz = "mat_plus x z"
  let ?yu = "mat_plus y u"
  from wf_mat_plus[OF wfx wfz] have wfxz: "wf_mat nr nc ?xz" .
  from wf_mat_plus[OF wfy wfu] have wfyu: "wf_mat nr nc ?yu" .
  show ?thesis 
  proof (simp only: mat_ge_index[OF wfxz wfyu], intro allI impI)
    fix i j
    assume i: "i < nc" and j: "j < nr"
    show "gt (mat_plus x z ! i ! j) (mat_plus y u ! i ! j)"
    proof (
	simp only: mat_plus_index[OF wfx wfz i j],
	simp only: mat_plus_index[OF wfy wfu i j],
        rule plus_gt_both_mono)
      show "gt (x ! i ! j) (y ! i ! j)" using gt1 i j mat_ge_index[OF wfx wfy] by auto
    next
      show "gt (z ! i ! j) (u ! i ! j)" using gt2 i j mat_ge_index[OF wfz wfu] by auto
    qed
  qed
qed


lemma mat_gt_arc_mult_mono: assumes gt1: "mat_gt_arc x y"
  and wfx: "wf_mat nr nc x"
  and wfy: "wf_mat nr nc y"
  and wfz: "wf_mat nc ncc z"
  shows "mat_gt_arc (mat_mult nr x z) (mat_mult nr y z)"
proof -
  let ?xz = "mat_mult nr x z"
  let ?yz = "mat_mult nr y z"
  from wf_mat_mult[OF wfx wfz] have wfxz: "wf_mat nr ncc ?xz" .
  from wf_mat_mult[OF wfy wfz] have wfyz: "wf_mat nr ncc ?yz" .
  show ?thesis 
  proof (simp only: mat_ge_index[OF wfxz wfyz], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    have wfxj : "wf_vec nc (row x j)" using wf_row[OF wfx j] .
    have wfyj : "wf_vec nc (row y j)" using wf_row[OF wfy j] .
    have wfzi : "wf_vec nc (col z i)" using wf_col[OF wfz i] .
    show "gt (?xz ! i ! j) (?yz ! i ! j)"
    proof (
        simp only: mat_mult_to_scalar[OF wfx wfz j i],
	simp only: mat_mult_to_scalar[OF wfy wfz j i],
	rule scalar_prod_left_mono[OF wfxj wfyj wfzi],
	simp only: vec_ge_index[OF wfxj wfyj],
	intro allI impI
      )
      fix k
      assume k: "k < nc"
      from gt1 mat_ge_index[OF wfx wfy, of gt] j k
      show "gt (row x j ! k) (row y j ! k)"
	by (
	  simp only: row_col[OF wfx j k],
	  simp only: row_col[OF wfy j k],
	  unfold col_def, auto)
    qed
  qed
qed

lemma mat0_leastI: assumes wf: "wf_mat nr nc m"
  shows "mat_gt_arc m (mat0 nr nc)"
proof (simp only: mat_ge_index[OF wf wf_mat0], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "gt (m ! i ! j) (mat0I 0 nr nc ! i ! j)"
    by (simp only: mat0_index[OF i j] zero_leastI)
qed

lemma mat0_leastII: 
  assumes gt: "mat_gt_arc (mat0 nr nc) m"
  and wf: "wf_mat nr nc m"
  shows "m = mat0 nr nc"
proof (simp only: mat_eq_index[OF wf wf_mat0], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "m ! i ! j = mat0 nr nc ! i ! j"
  proof (simp only: mat0_index[OF i j], rule zero_leastII)
    show "gt 0 (m ! i ! j)" using i j gt mat_ge_index[OF wf_mat0 wf, of gt 0] mat0_index[OF i j, of "0 :: 'a"] by force
  qed
qed


lemma mat0_leastIII: assumes wf: "wf_mat nr nc m"
  shows "mat_ge m ((mat0 nr nc) :: 'a mat)"
proof (simp only: mat_ge_index[OF wf wf_mat0], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "(m ! i ! j) \<ge> ((mat0 nr nc :: 'a mat) ! i ! j)"
    by (simp only: mat0_index[OF i j] zero_leastIII)
qed

lemma mat1_gt_arc_mat0: "mat_gt_arc (mat1 n) (mat0 n n)"
proof (simp only: mat_ge_index[OF wf_mat1 wf_mat0], intro allI impI)
  fix i j
  assume i: "i < n" and j: "j < n"
  show "gt (mat1 n ! i ! j) (mat0 n n ! i ! j)"
    by (simp only: mat0_index[OF i j],
      rule zero_leastI)
qed
end 

locale SN_both_mono_ordered_semiring_1 = both_mono_ordered_semiring_1 +
  fixes arc_pos :: "'a \<Rightarrow> bool"
  assumes SN: "SN {(x,y) . arc_pos y \<and> gt x y}"
  and arc_pos_plus: "arc_pos x \<Longrightarrow> arc_pos (x + y)"
  and arc_pos_mult: "\<lbrakk>arc_pos x; arc_pos y\<rbrakk> \<Longrightarrow> arc_pos (x * y)"
  and arc_pos_one: "arc_pos 1"
begin



abbreviation mat_arc_pos :: "'a mat \<Rightarrow> bool"
where "mat_arc_pos \<equiv> mat_arc_posI arc_pos"


lemma mat_gt_arc_SN: assumes n_pos: "n > 0" shows
  "SN {(x, y). wf_mat n n x \<and> wf_mat n n y \<and> mat_arc_pos y \<and> mat_gt_arc x y}" (is "SN ?rel")
proof (rule ccontr)
  assume "\<not> SN ?rel"
  from this obtain f and x where "f (0 :: nat) = x" and steps: "\<forall> i. (f i, f (Suc i)) \<in> ?rel" unfolding SN_defs by blast
  hence pos: "\<forall> i. arc_pos (f (Suc i) ! 0 ! 0)" unfolding mat_arc_posI_def by blast
  have gt: "\<forall> i. gt (f i ! 0 ! 0)  (f (Suc i) ! 0 ! 0)"
  proof
    fix i
    from steps have wf1: "wf_mat n n (f i)" and wf2: "wf_mat n n (f (Suc i))" and gt: "mat_gt_arc (f i) (f (Suc i))" by auto
    show "gt (f i ! 0 ! 0)  (f (Suc i) ! 0 ! 0)" using mat_ge_index[OF wf1 wf2, of gt] mat0_index[OF n_pos n_pos, of 0] gt n_pos by force
  qed
  from pos gt SN show False unfolding SN_defs by force
qed



lemma mat_arc_pos_plus: assumes n_pos: "n > 0" 
  and wf1: "wf_mat n n m1"
  and wf2: "wf_mat n n m2"
  and arc_pos: "mat_arc_pos m1"
  shows "mat_arc_pos (mat_plus m1 m2)"
proof -
  from n_pos wf1 obtain v1 mm1 where m1: "m1 = v1 # mm1" unfolding wf_mat_def by (cases m1, auto)
  from n_pos wf2 obtain v2 mm2 where m2: "m2 = v2 # mm2" unfolding wf_mat_def by (cases m2, auto)	
  from n_pos wf1 m1 obtain a1 vv1 where v1: "v1 = a1 # vv1" unfolding wf_mat_def by (cases v1, auto simp: wf_vec_def)
  from n_pos wf2 m2 obtain a2 vv2 where v2: "v2 = a2 # vv2" unfolding wf_mat_def by (cases v2, auto simp: wf_vec_def)
  from m1 v1 arc_pos have "arc_pos a1" unfolding mat_arc_posI_def by simp
  hence "arc_pos (plus a1 a2)" by (rule arc_pos_plus)
  with m1 v1 m2 v2 show ?thesis unfolding mat_arc_posI_def mat_plusI_def vec_plusI_def by simp
qed

lemma mat_arc_pos_mult: assumes n_pos: "n > 0" 
  and wf1: "wf_mat n n m1"
  and wf2: "wf_mat n n m2"
  and ap1: "mat_arc_pos m1"
  and ap2: "mat_arc_pos m2"
  shows "mat_arc_pos (mat_mult n m1 m2)"
proof -
  from n_pos wf1 obtain v1 mm1 where m1: "m1 = v1 # mm1" unfolding wf_mat_def by (cases m1, auto)
  from n_pos wf2 obtain v2 mm2 where m2: "m2 = v2 # mm2" unfolding wf_mat_def by (cases m2, auto)	
  from n_pos wf1 m1 obtain a1 vv1 where v1: "v1 = a1 # vv1" unfolding wf_mat_def by (cases v1, auto simp: wf_vec_def)
  from n_pos wf2 m2 obtain a2 vv2 where v2: "v2 = a2 # vv2" unfolding wf_mat_def by (cases v2, auto simp: wf_vec_def)
  from m1 v1 ap1 have a1: "arc_pos a1" unfolding mat_arc_posI_def by simp
  from m2 v2 ap2 have a2: "arc_pos a2" unfolding mat_arc_posI_def by simp
  from a1 a2 have prod: "arc_pos (a1 * a2)" by (rule arc_pos_mult)
  show ?thesis unfolding mat_arc_posI_def 
    by (simp only: mat_mult_to_scalar[OF wf1 wf2 n_pos n_pos],
    simp add: m1 v1 m2 v2 col_def row_def scalar_prod_cons,
    rule arc_pos_plus[OF prod])
qed

lemma mat_arc_pos_mat1: assumes n_pos: "n > 0" shows "mat_arc_pos (mat1 n)"
unfolding mat1I_def mat_arc_posI_def 
using n_pos
by (auto simp: vec1I_def arc_pos_one)
end


declare wf_vec0[simp del] wf_mat0[simp del] vec0_plus[simp del] plus_vec0[simp del] plus_mat0[simp del]

end
